"""
Norm Scorer
-----------

Norm anomaly score (of given order) [1]_.

References
----------
.. [1] https://en.wikipedia.org/wiki/Norm_(mathematics)
"""

import numpy as np

from darts.ad.scorers.scorers import NonFittableAnomalyScorer
from darts.logging import raise_if_not
from darts.timeseries import TimeSeries


class NormScorer(NonFittableAnomalyScorer):
    def __init__(self, ord=None, component_wise: bool = False) -> None:
        """
        Returns the elementwise norm of a given order between two series' values.

        If `component_wise` is False, the norm is computed between vectors
        made of the series' components (one norm per timestamp).

        If `component_wise` is True, for any `ord` this effectively amounts to computing the absolute
        value of the difference.

        The scoring function expects two series.

        If the two series are multivariate of width `w`:

        * if `component_wise` is set to False: it returns a univariate series (width=1).
        * if `component_wise` is set to True: it returns a multivariate series of width `w`.

        If the two series are univariate, it returns a univariate series regardless of the parameter
        `component_wise`.

        Parameters
        ----------
        ord
            Order of the norm. Options are listed under 'Notes' at:
            <https://numpy.org/doc/stable/reference/generated/numpy.linalg.norm.html>.
            Default: None
        component_wise
            Whether to compare components of the two series in isolation (True), or jointly (False).
            Default: False
        """

        raise_if_not(
            type(component_wise) is bool,
            f"`component_wise` must be Boolean, found type: {type(component_wise)}.",
        )

        self.ord = ord
        self.component_wise = component_wise
        super().__init__(univariate_scorer=(not component_wise), window=1)

    def __str__(self):
        return f"Norm (ord={self.ord})"

    def _score_core_from_prediction(
        self,
        actual_series: TimeSeries,
        pred_series: TimeSeries,
    ) -> TimeSeries:

        self._assert_deterministic(actual_series, "actual_series")
        self._assert_deterministic(pred_series, "pred_series")

        diff = actual_series - pred_series

        if self.component_wise:
            return diff.map(lambda x: np.abs(x))

        else:
            diff_np = diff.all_values(copy=False)

            return TimeSeries.from_times_and_values(
                diff.time_index, np.linalg.norm(diff_np, ord=self.ord, axis=1)
            )



import numpy as np
import pandas as pd
import pytest

from darts import TimeSeries
from darts.dataprocessing.encoders.encoder_base import (
    CovariatesIndexGenerator,
    FutureCovariatesIndexGenerator,
    PastCovariatesIndexGenerator,
)
from darts.logging import get_logger
from darts.tests.base_test_class import DartsBaseTestClass
from darts.utils import timeseries_generation as tg

logger = get_logger(__name__)


class CovariatesIndexGeneratorTestCase(DartsBaseTestClass):
    n_target = 24
    target_time = tg.linear_timeseries(length=n_target, freq="MS")
    cov_time_train = tg.datetime_attribute_timeseries(
        target_time, attribute="month", cyclic=True
    )
    cov_time_train_short = cov_time_train[1:]

    target_int = tg.linear_timeseries(length=n_target, start=2)
    cov_int_train = target_int
    cov_int_train_short = cov_int_train[1:]

    input_chunk_length = 12
    output_chunk_length = 6
    n_short = 6
    n_long = 8

    # pd.DatetimeIndex
    # expected covariates for inference dataset for n <= output_chunk_length
    cov_time_inf_short = TimeSeries.from_times_and_values(
        tg.generate_index(
            start=target_time.start_time(),
            length=n_target + n_short,
            freq=target_time.freq,
        ),
        np.arange(n_target + n_short),
    )
    # expected covariates for inference dataset for n > output_chunk_length
    cov_time_inf_long = TimeSeries.from_times_and_values(
        tg.generate_index(
            start=target_time.start_time(),
            length=n_target + n_long,
            freq=target_time.freq,
        ),
        np.arange(n_target + n_long),
    )

    # integer index
    # excpected covariates for inference dataset for n <= output_chunk_length
    cov_int_inf_short = TimeSeries.from_times_and_values(
        tg.generate_index(
            start=target_int.start_time(),
            length=n_target + n_short,
            freq=target_int.freq,
        ),
        np.arange(n_target + n_short),
    )
    # excpected covariates for inference dataset for n > output_chunk_length
    cov_int_inf_long = TimeSeries.from_times_and_values(
        tg.generate_index(
            start=target_int.start_time(),
            length=n_target + n_long,
            freq=target_int.freq,
        ),
        np.arange(n_target + n_long),
    )

    def helper_test_index_types(self, ig: CovariatesIndexGenerator):
        """test the index type of generated index"""
        # pd.DatetimeIndex
        idx, _ = ig.generate_train_idx(self.target_time, self.cov_time_train)
        self.assertTrue(isinstance(idx, pd.DatetimeIndex))
        idx, _ = ig.generate_inference_idx(
            self.n_short, self.target_time, self.cov_time_inf_short
        )
        self.assertTrue(isinstance(idx, pd.DatetimeIndex))
        idx, _ = ig.generate_train_idx(self.target_time, None)
        self.assertTrue(isinstance(idx, pd.DatetimeIndex))

        # pd.RangeIndex
        idx, _ = ig.generate_train_idx(self.target_int, self.cov_int_train)
        self.assertTrue(isinstance(idx, pd.RangeIndex))
        idx, _ = ig.generate_inference_idx(
            self.n_short, self.target_int, self.cov_int_inf_short
        )
        self.assertTrue(isinstance(idx, pd.RangeIndex))
        idx, _ = ig.generate_train_idx(self.target_int, None)
        self.assertTrue(isinstance(idx, pd.RangeIndex))

    def helper_test_index_generator_train(self, ig: CovariatesIndexGenerator):
        """
        If covariates are given, the index generators should return the covariate series' index.
        If covariates are not given, the index generators should return the target series' index.
        """
        # pd.DatetimeIndex
        # generated index must be equal to input covariate index
        idx, _ = ig.generate_train_idx(self.target_time, self.cov_time_train)
        self.assertTrue(idx.equals(self.cov_time_train.time_index))
        # generated index must be equal to input covariate index
        idx, _ = ig.generate_train_idx(self.target_time, self.cov_time_train_short)
        self.assertTrue(idx.equals(self.cov_time_train_short.time_index))
        # generated index must be equal to input target index when no covariates are defined
        idx, _ = ig.generate_train_idx(self.target_time, None)
        self.assertEqual(idx[0], self.target_time.start_time())
        if isinstance(ig, PastCovariatesIndexGenerator):
            self.assertEqual(
                idx[-1],
                self.target_time.end_time()
                - self.output_chunk_length * self.target_time.freq,
            )
        else:
            self.assertEqual(idx[-1], self.target_time.end_time())

        # integer index
        # generated index must be equal to input covariate index
        idx, _ = ig.generate_train_idx(self.target_int, self.cov_int_train)
        self.assertTrue(idx.equals(self.cov_int_train.time_index))
        # generated index must be equal to input covariate index
        idx, _ = ig.generate_train_idx(self.target_int, self.cov_int_train_short)
        self.assertTrue(idx.equals(self.cov_int_train_short.time_index))
        # generated index must be equal to input target index when no covariates are defined
        idx, _ = ig.generate_train_idx(self.target_int, None)
        self.assertEqual(idx[0], self.target_int.start_time())
        if isinstance(ig, PastCovariatesIndexGenerator):
            self.assertEqual(
                idx[-1],
                self.target_int.end_time()
                - self.output_chunk_length * self.target_int.freq,
            )
        else:
            self.assertEqual(idx[-1], self.target_int.end_time())

    def helper_test_index_generator_inference(self, ig, is_past=False):
        """
        For prediction (`n` is given) with past covariates we have to distinguish between two cases:
        1)  if past covariates are given, we can use them as reference
        2)  if past covariates are missing, we need to generate a time index that starts `input_chunk_length`
            before the end of `target` and ends `max(0, n - output_chunk_length)` after the end of `target`

        For prediction (`n` is given) with future covariates we have to distinguish between two cases:
        1)  if future covariates are given, we can use them as reference
        2)  if future covariates are missing, we need to generate a time index that starts `input_chunk_length`
            before the end of `target` and ends `max(n, output_chunk_length)` after the end of `target`
        """

        # check generated inference index without passing covariates when n <= output_chunk_length
        idx, _ = ig.generate_inference_idx(self.n_short, self.target_time, None)
        if is_past:
            n_out = self.input_chunk_length
            last_idx = self.target_time.end_time()
        else:
            n_out = self.input_chunk_length + self.output_chunk_length
            last_idx = self.cov_time_inf_short.end_time()

        self.assertTrue(len(idx) == n_out)
        self.assertTrue(idx[-1] == last_idx)

        # check generated inference index without passing covariates when n > output_chunk_length
        idx, _ = ig.generate_inference_idx(self.n_long, self.target_time, None)
        if is_past:
            n_out = self.input_chunk_length + self.n_long - self.output_chunk_length
            last_idx = (
                self.target_time.end_time()
                + (self.n_long - self.output_chunk_length) * self.target_time.freq
            )
        else:
            n_out = self.input_chunk_length + self.n_long
            last_idx = self.cov_time_inf_long.end_time()

        self.assertTrue(len(idx) == n_out)
        self.assertTrue(idx[-1] == last_idx)

        idx, _ = ig.generate_inference_idx(
            self.n_short, self.target_time, self.cov_time_inf_short
        )
        self.assertTrue(idx.equals(self.cov_time_inf_short.time_index))
        idx, _ = ig.generate_inference_idx(
            self.n_long, self.target_time, self.cov_time_inf_long
        )
        self.assertTrue(idx.equals(self.cov_time_inf_long.time_index))
        idx, _ = ig.generate_inference_idx(
            self.n_short, self.target_int, self.cov_int_inf_short
        )
        self.assertTrue(idx.equals(self.cov_int_inf_short.time_index))
        idx, _ = ig.generate_inference_idx(
            self.n_long, self.target_int, self.cov_int_inf_long
        )
        self.assertTrue(idx.equals(self.cov_int_inf_long.time_index))

    def helper_test_index_generator_creation(self, ig_cls, is_past=False):
        # invalid parameter sets
        with pytest.raises(ValueError):
            _ = ig_cls(
                output_chunk_length=3,
            )

        with pytest.raises(ValueError):
            _ = ig_cls(
                output_chunk_length=3,
                lags_covariates=[-1],
            )

        # valid parameter sets
        # TorchForecastingModel scenario
        _ = ig_cls(
            input_chunk_length=3,
            output_chunk_length=3,
        )
        # RegressionModel scenario
        _ = ig_cls(
            input_chunk_length=3,
            output_chunk_length=3,
            lags_covariates=[-1],
        )
        # LocalForecastingModel scenario, or model agnostic (only supported by FutureCovariatesIndexGenerator)
        if not is_past:
            _ = ig_cls()

    def test_past_index_generator_creation(self):
        # test parameter scenarios
        self.helper_test_index_generator_creation(
            ig_cls=PastCovariatesIndexGenerator, is_past=True
        )

        # ==> test failures
        # one lag is >= 0 (not possible for past covariates)
        with pytest.raises(ValueError):
            _ = PastCovariatesIndexGenerator(
                1,
                1,
                lags_covariates=[-1, 1],
            )
        with pytest.raises(ValueError):
            _ = PastCovariatesIndexGenerator(
                1,
                1,
                lags_covariates=[0, -1],
            )

        min_lag, max_lag = -2, -1
        ig = PastCovariatesIndexGenerator(
            1,
            1,
            lags_covariates=[min_lag, max_lag],
        )
        self.assertEqual(ig.shift_start, min_lag + 1)
        self.assertEqual(ig.shift_end, max_lag + 1)

        min_lag, max_lag = -1, -1
        ig = PastCovariatesIndexGenerator(
            1,
            1,
            lags_covariates=[min_lag, max_lag],
        )
        self.assertEqual(ig.shift_start, min_lag + 1)
        self.assertEqual(ig.shift_end, max_lag + 1)

        # check that min/max lags are extracted from list of lags
        min_lag, max_lag = -10, -3
        ig = PastCovariatesIndexGenerator(
            1,
            1,
            lags_covariates=[-5, min_lag, max_lag, -4],
        )
        self.assertEqual(ig.shift_start, min_lag + 1)
        self.assertEqual(ig.shift_end, max_lag + 1)

    def test_future_index_generator_creation(self):
        # test parameter scenarios
        self.helper_test_index_generator_creation(
            ig_cls=FutureCovariatesIndexGenerator, is_past=False
        )

        # future covariates index generator (ig) can technically be used like a past covariates ig
        min_lag, max_lag = -2, -1
        ig = FutureCovariatesIndexGenerator(
            1,
            1,
            lags_covariates=[min_lag, max_lag],
        )
        self.assertEqual(ig.shift_start, min_lag + 1)
        self.assertEqual(ig.shift_end, max_lag + 1)

        min_lag, max_lag = -1, -1
        ig = FutureCovariatesIndexGenerator(
            1,
            1,
            lags_covariates=[min_lag, max_lag],
        )
        self.assertEqual(ig.shift_start, min_lag + 1)
        self.assertEqual(ig.shift_end, max_lag + 1)

        # different to past covariates ig, future ig can take positive and negative lags
        min_lag, max_lag = -2, 1
        ig = FutureCovariatesIndexGenerator(
            1,
            1,
            lags_covariates=[min_lag, max_lag],
        )
        self.assertEqual(ig.shift_start, min_lag + 1)
        # when `max_lag` >= 0, we add one step to `shift_end`, as future lags start at 0 meaning first prediction step
        self.assertEqual(ig.shift_end, max_lag + 1)

        # check that min/max lags are extracted from list of lags
        min_lag, max_lag = -10, 5
        ig = FutureCovariatesIndexGenerator(
            1,
            1,
            lags_covariates=[-5, min_lag, max_lag, -1],
        )
        self.assertEqual(ig.shift_start, min_lag + 1)
        self.assertEqual(ig.shift_end, max_lag + 1)

    def test_past_index_generator(self):
        ig = PastCovariatesIndexGenerator(
            self.input_chunk_length, self.output_chunk_length
        )
        self.helper_test_index_types(ig)
        self.helper_test_index_generator_train(ig)
        self.helper_test_index_generator_inference(ig, is_past=True)

    def test_past_index_generator_with_lags(self):
        icl = self.input_chunk_length
        ocl = self.output_chunk_length
        target = self.target_time

        def test_routine_train(
            self, icl, ocl, min_lag, max_lag, start_expected, end_expected
        ):
            idxg = PastCovariatesIndexGenerator(
                icl,
                ocl,
                lags_covariates=[min_lag, max_lag],
            )
            idx, _ = idxg.generate_train_idx(target, None)
            self.assertEqual(idx[0], pd.Timestamp(start_expected))
            self.assertEqual(idx[-1], pd.Timestamp(end_expected))
            # check case 0: we give covariates, index will always be the covariate time index
            idx, _ = idxg.generate_train_idx(target, self.cov_time_train)
            self.assertTrue(idx.equals(self.cov_time_train.time_index))
            return idxg

        def test_routine_inf(self, idxg, n, start_expected, end_expected):
            idx, _ = idxg.generate_inference_idx(n, target, None)
            self.assertEqual(idx[0], pd.Timestamp(start_expected))
            self.assertEqual(idx[-1], pd.Timestamp(end_expected))
            # check case 0: we give covariates, index will always be the covariate time index
            idx, _ = idxg.generate_inference_idx(n, target, self.cov_time_inf_short)
            self.assertTrue(idx.equals(self.cov_time_inf_short.time_index))

        # lags are required for RegressionModels
        # case 1: abs(min_lags) == icl and abs(max_lag) == -1:
        # will give identical results as without setting lags
        min_lag = -12  # = -icl
        max_lag = -1
        expected_start = "2000-01-01"
        expected_end = "2001-06-01"
        ig = test_routine_train(
            self, icl, ocl, min_lag, max_lag, expected_start, expected_end
        )
        self.helper_test_index_types(ig)
        self.helper_test_index_generator_train(ig)
        self.helper_test_index_generator_inference(ig, is_past=True)
        # check inference for n <= ocl
        expected_start = "2001-01-01"
        expected_end = "2001-12-01"
        test_routine_inf(self, ig, 1, expected_start, expected_end)
        test_routine_inf(self, ig, ocl, expected_start, expected_end)
        # check inference for n > ocl
        test_routine_inf(self, ig, ocl + 1, expected_start, "2002-01-01")

        # case 2: abs(min_lag) < icl and abs(max_lag) == -1:
        # the start time of covariates begins before target start
        min_lag, max_lag = -11, -1
        expected_start = "2000-02-01"
        expected_end = "2001-06-01"
        ig = test_routine_train(
            self, icl, ocl, min_lag, max_lag, expected_start, expected_end
        )
        # check inference for n <= ocl
        expected_start = "2001-02-01"
        expected_end = "2001-12-01"
        test_routine_inf(self, ig, 1, expected_start, expected_end)
        test_routine_inf(self, ig, ocl, expected_start, expected_end)
        # check inference for n > ocl
        test_routine_inf(self, ig, ocl + 1, expected_start, "2002-01-01")

        # case 3: abs(min_lag) > icl and abs(max_lag) == -1:
        # the start time of covariates begins before target start
        min_lag, max_lag = -13, -1
        expected_start = "1999-12-01"
        expected_end = "2001-06-01"
        ig = test_routine_train(
            self, icl, ocl, min_lag, max_lag, expected_start, expected_end
        )
        # check inference for n <= ocl
        expected_start = "2000-12-01"
        expected_end = "2001-12-01"
        test_routine_inf(self, ig, 1, expected_start, expected_end)
        test_routine_inf(self, ig, ocl, expected_start, expected_end)
        # check inference for n > ocl
        test_routine_inf(self, ig, ocl + 1, expected_start, "2002-01-01")

        # case 4: abs(min_lag) > icl and abs(max_lag) > -1:
        # the start time of covariates begins before target start
        min_lag, max_lag = -13, -2
        expected_start = "1999-12-01"
        expected_end = "2001-05-01"
        ig = test_routine_train(
            self, icl, ocl, min_lag, max_lag, expected_start, expected_end
        )
        # check inference for n <= ocl
        expected_start = "2000-12-01"
        expected_end = "2001-11-01"
        test_routine_inf(self, ig, 1, expected_start, expected_end)
        test_routine_inf(self, ig, ocl, expected_start, expected_end)
        # check inference for n > ocl
        test_routine_inf(self, ig, ocl + 1, expected_start, "2001-12-01")

    def test_future_index_generator(self):
        ig = FutureCovariatesIndexGenerator(
            self.input_chunk_length, self.output_chunk_length
        )
        self.helper_test_index_types(ig)
        self.helper_test_index_generator_train(ig)
        self.helper_test_index_generator_inference(ig, is_past=False)

    def test_future_index_generator_with_lags(self):
        icl = self.input_chunk_length
        ocl = self.output_chunk_length
        target = self.target_time

        def test_routine_train(
            self, icl, ocl, min_lag, max_lag, start_expected, end_expected
        ):
            idxg = FutureCovariatesIndexGenerator(
                icl,
                ocl,
                lags_covariates=[min_lag, max_lag],
            )
            idx, _ = idxg.generate_train_idx(target, None)
            self.assertEqual(idx[0], pd.Timestamp(start_expected))
            self.assertEqual(idx[-1], pd.Timestamp(end_expected))
            # check case 0: we give covariates, index will always be the covariate time index
            idx, _ = idxg.generate_train_idx(target, self.cov_time_train)
            self.assertTrue(idx.equals(self.cov_time_train.time_index))
            return idxg

        def test_routine_inf(self, idxg, n, start_expected, end_expected):
            idx, _ = idxg.generate_inference_idx(n, target, None)
            self.assertTrue(idx[0], pd.Timestamp(start_expected))
            self.assertTrue(idx[-1], pd.Timestamp(end_expected))
            # check case 0: we give covariates, index will always be the covariate time index
            idx, _ = idxg.generate_inference_idx(n, target, self.cov_time_inf_short)
            self.assertTrue(idx.equals(self.cov_time_inf_short.time_index))

        # INFO: test cases 1, 2, and 3 only have lags in the past which yields identical results as using a
        # PastCovariatesIndexGenerator
        # case 1: abs(min_lag) < icl and abs(max_lag) == -1:
        # the start time of covariates begins before target start
        min_lag, max_lag = -11, -1
        expected_start = "2000-02-01"
        expected_end = "2001-06-01"
        ig = test_routine_train(
            self, icl, ocl, min_lag, max_lag, expected_start, expected_end
        )
        # check inference for n <= ocl
        expected_start = "2001-02-01"
        expected_end = "2001-12-01"
        test_routine_inf(self, ig, 1, expected_start, expected_end)
        test_routine_inf(self, ig, ocl, expected_start, expected_end)
        # check inference for n > ocl
        test_routine_inf(self, ig, ocl + 1, expected_start, "2002-01-01")

        # case 2: abs(min_lag) > icl and abs(max_lag) == -1:
        # the start time of covariates begins before target start
        min_lag, max_lag = -13, -1
        expected_start = "1999-12-01"
        expected_end = "2001-06-01"
        ig = test_routine_train(
            self, icl, ocl, min_lag, max_lag, expected_start, expected_end
        )
        # check inference for n <= ocl
        expected_start = "2000-12-01"
        expected_end = "2001-12-01"
        test_routine_inf(self, ig, 1, expected_start, expected_end)
        test_routine_inf(self, ig, ocl, expected_start, expected_end)
        # check inference for n > ocl
        test_routine_inf(self, ig, ocl + 1, expected_start, "2002-01-01")

        # case 3: abs(min_lag) > icl and abs(max_lag) > -1:
        # the start time of covariates begins before target start
        min_lag, max_lag = -13, -2
        expected_start = "1999-12-01"
        expected_end = "2001-05-01"
        ig = test_routine_train(
            self, icl, ocl, min_lag, max_lag, expected_start, expected_end
        )
        # check inference for n <= ocl
        expected_start = "2000-12-01"
        expected_end = "2001-11-01"
        test_routine_inf(self, ig, 1, expected_start, expected_end)
        test_routine_inf(self, ig, ocl, expected_start, expected_end)
        # check inference for n > ocl
        test_routine_inf(self, ig, ocl + 1, expected_start, "2001-12-01")

        # INFO: the following test cases have lags in the future which is different to PastCovariatesIndexGenerator
        # case 4: abs(min_lags) == icl and max_lag == (ocl - 1):
        # will give identical results as without setting lags
        min_lag = -12  # -icl
        max_lag = 5  # (ocl - 1)
        expected_start = "2000-01-01"
        expected_end = "2001-12-01"
        ig = test_routine_train(
            self, icl, ocl, min_lag, max_lag, expected_start, expected_end
        )
        self.helper_test_index_types(ig)
        self.helper_test_index_generator_train(ig)
        self.helper_test_index_generator_inference(ig, is_past=False)
        # check inference for n <= ocl
        expected_start = "2001-01-01"
        expected_end = "2002-06-01"
        test_routine_inf(self, ig, 1, expected_start, expected_end)
        test_routine_inf(self, ig, ocl, expected_start, expected_end)
        # check inference for n > ocl
        test_routine_inf(self, ig, ocl + 1, expected_start, "2002-07-01")

        # case 5: abs(min_lag) == icl and max_lag < (ocl - 1)
        # the end of covariates should be one time step after beginning of last output chunk with max_lag = 0
        min_lag, max_lag = -12, 0
        expected_start = "2000-01-01"
        expected_end = "2001-07-01"
        ig = test_routine_train(
            self, icl, ocl, min_lag, max_lag, expected_start, expected_end
        )
        expected_start = "2001-01-01"
        expected_end = "2002-01-01"
        test_routine_inf(self, ig, 1, expected_start, expected_end)
        test_routine_inf(self, ig, ocl, expected_start, expected_end)
        # check inference for n > ocl
        test_routine_inf(self, ig, ocl + 1, expected_start, "2002-02-01")

        # case 6: abs(min_lag) == icl and max_lag > (ocl - 1)
        # the end of covariates is after the end of target series with max_lag = (ocl - 1) + 1
        min_lag, max_lag = -12, 17
        expected_start = "2000-01-01"
        expected_end = "2002-12-01"
        ig = test_routine_train(
            self, icl, ocl, min_lag, max_lag, expected_start, expected_end
        )
        expected_start = "2001-01-01"
        expected_end = "2003-01-01"
        test_routine_inf(self, ig, 1, expected_start, expected_end)
        test_routine_inf(self, ig, ocl, expected_start, expected_end)
        # check inference for n > ocl
        test_routine_inf(self, ig, ocl + 1, expected_start, "2003-02-01")

        # case 7: min_lag >= 0 and max_lag <= (ocl - 1)
        # only future part of future covariates (no historical part)
        min_lag, max_lag = 0, 2
        expected_start = "2001-01-01"
        expected_end = "2001-09-01"
        ig = test_routine_train(
            self, icl, ocl, min_lag, max_lag, expected_start, expected_end
        )
        expected_start = "2002-01-01"
        expected_end = "2002-03-01"
        test_routine_inf(self, ig, 1, expected_start, expected_end)
        test_routine_inf(self, ig, ocl, expected_start, expected_end)
        # check inference for n > ocl
        test_routine_inf(self, ig, ocl + 1, expected_start, "2002-04-01")

        # case 8: min_lag >= 0 and max_lag > (ocl - 1)
        # only future part of future covariates (no historical part)
        min_lag, max_lag = 0, 17
        expected_start = "2001-01-01"
        expected_end = "2002-12-01"
        ig = test_routine_train(
            self, icl, ocl, min_lag, max_lag, expected_start, expected_end
        )
        expected_start = "2002-01-01"
        expected_end = "2003-01-01"
        test_routine_inf(self, ig, 1, expected_start, expected_end)
        test_routine_inf(self, ig, ocl, expected_start, expected_end)
        # check inference for n > ocl
        test_routine_inf(self, ig, ocl + 1, expected_start, "2003-02-01")

        # case 9: abs(min_lag) > icl and max_lag > (ocl - 1)
        min_lag, max_lag = -13, 17
        expected_start = "1999-12-01"
        expected_end = "2002-12-01"
        ig = test_routine_train(
            self, icl, ocl, min_lag, max_lag, expected_start, expected_end
        )
        expected_start = "2000-12-01"
        expected_end = "2002-03-01"
        test_routine_inf(self, ig, 1, expected_start, expected_end)
        test_routine_inf(self, ig, ocl, expected_start, expected_end)
        # check inference for n > ocl
        test_routine_inf(self, ig, ocl + 1, expected_start, "2002-04-01")

    def test_future_index_generator_local(self):
        # test model agnostic scenario (also for LocalForecastingModels)
        freq = self.target_time.freq
        target = self.target_time

        idxg = FutureCovariatesIndexGenerator()
        idx, _ = idxg.generate_train_idx(target=target, covariates=None)
        self.assertTrue(idx.equals(target.time_index))
        idx, _ = idxg.generate_train_idx(target=target, covariates=self.cov_time_train)
        self.assertTrue(idx.equals(self.cov_time_train.time_index))

        n = 10
        idx, _ = idxg.generate_inference_idx(n=n, target=target, covariates=None)
        self.assertEqual(idx.freq, freq)
        self.assertEqual(idx[0], target.end_time() + 1 * freq)
        self.assertEqual(idx[-1], target.end_time() + n * freq)

        idx, _ = idxg.generate_inference_idx(
            n=n, target=target, covariates=self.cov_int_inf_short
        )
        self.assertTrue(idx.equals(self.cov_int_inf_short.time_index))



"""
Fittable Data Transformer Base Class
------------------------------------
"""

from abc import abstractmethod
from typing import Any, Generator, List, Mapping, Optional, Sequence, Union

import numpy as np

from darts import TimeSeries
from darts.logging import get_logger, raise_if, raise_if_not
from darts.utils import _build_tqdm_iterator, _parallel_apply

from .base_data_transformer import BaseDataTransformer

logger = get_logger(__name__)


class FittableDataTransformer(BaseDataTransformer):
    def __init__(
        self,
        name: str = "FittableDataTransformer",
        n_jobs: int = 1,
        verbose: bool = False,
        parallel_params: Union[bool, Sequence[str]] = False,
        mask_components: bool = True,
        global_fit: bool = False,
    ):
        """Base class for fittable transformers.

        All the deriving classes have to implement the static methods
        :func:`ts_transform()` and :func:`ts_fit()`. The fitting and transformation functions must
        be passed during the transformer's initialization. This class takes care of parallelizing
        operations involving multiple ``TimeSeries`` when possible.

        Parameters
        ----------
        name
            The data transformer's name
        n_jobs
            The number of jobs to run in parallel. Parallel jobs are created only when a `Sequence[TimeSeries]` is
            passed as input to a method, parallelising operations regarding different `TimeSeries`. Defaults to `1`
            (sequential). Setting the parameter to `-1` means using all the available processors.
            Note: for a small amount of data, the parallelisation overhead could end up increasing the total
            required amount of time.
        verbose
            Optionally, whether to print operations progress
        parallel_params
            Optionally, specifies which fixed parameters (i.e. the attributes initialized in the child-most
            class's `__init__`) take on different values for different parallel jobs. Fixed parameters specified
            by `parallel_params` are assumed to be a `Sequence` of values that should be used for that parameter
            in each parallel job; the length of this `Sequence` should equal the number of parallel jobs. If
            `parallel_params=True`, every fixed parameter will take on a different value for each
            parallel job. If `parallel_params=False`, every fixed parameter will take on the same value for
            each parallel job. If `parallel_params` is a `Sequence` of fixed attribute names, only those
            attribute names specified will take on different values between different parallel jobs.
        mask_components
            Optionally, whether or not to automatically apply any provided `component_mask`s to the
            `TimeSeries` inputs passed to `transform`, `fit`, `inverse_transform`, or `fit_transform`.
            If `True`, any specified `component_mask` will be applied to each input timeseries
            before passing them to the called method; the masked components will also be automatically
            'unmasked' in the returned `TimeSeries`. If `False`, then `component_mask` (if provided) will
            be passed as a keyword argument, but won't automatically be applied to the input timeseries.
            See `apply_component_mask` method of `BaseDataTransformer` for further details.
        global_fit
            Optionally, whether all of the `TimeSeries` passed to the `fit()` method should be used to fit
            a *single* set of parameters, or if a different set of parameters should be independently fitted
            to each provided `TimeSeries`. If `True`, then a `Sequence[TimeSeries]` is passed to `ts_fit`
            and a single set of parameters is fitted using all of the provided `TimeSeries`. If `False`, then
            each `TimeSeries` is individually passed to `ts_fit`, and a different set of fitted parameters
            if yielded for each of these fitting operations. See `ts_fit` for further details.

        Notes
        -----
        If `global_fit` is `False` and `fit` is called with a `Sequence` containing `n` different `TimeSeries`,
        then `n` sets of parameters will be fitted. When `transform` and/or `inverse_transform` is subsequently
        called with a `Series[TimeSeries]`, the `i`th set of fitted parameter values will be passed to
        `ts_transform`/`ts_inverse_transform` to transform the `i`th `TimeSeries` in this sequence. Conversely,
        if `global_fit` is `True`, then only a single set of fitted values will be produced when `fit` is
        provided with a `Sequence[TimeSeries]`. Consequently, if a `Sequence[TimeSeries]` is then passed to
        `transform`/`inverse_transform`, each of these `TimeSeries` will be transformed using the exact same set
        of fitted parameters.

        Note that if an invertible *and* fittable data transformer is to be globally fitted, the data transformer
        class should first inherit from `FittableDataTransformer` and then from `InveritibleDataTransformer`. In
        other words, `MyTransformer(FittableDataTransformer, InveritibleDataTransformer)` is correct, but
        `MyTransformer(InveritibleDataTransformer, FittableDataTransformer)` is **not**. If this is not implemented
        correctly, then the `global_fit` parameter will not be correctly passed to `FittableDataTransformer`'s
        constructor.

        The :func:`ts_transform()` and :func:`ts_fit()` methods are designed to be static methods instead of instance
        methods to allow an efficient parallelisation also when the scaler instance is storing a non-negligible
        amount of data. Using instance methods would imply copying the instance's data through multiple processes, which
        can easily introduce a bottleneck and nullify parallelisation benefits.

        Example
        --------
        >>> from darts.dataprocessing.transformers import FittableDataTransformer
        >>> from darts.utils.timeseries_generation import linear_timeseries
        >>>
        >>> class SimpleRangeScaler(FittableDataTransformer):
        >>>
        >>>     def __init__(self, scale, position):
        >>>         self._scale = scale
        >>>         self._position = position
        >>>         super().__init__()
        >>>
        >>>     @staticmethod
        >>>     def ts_transform(series, params):
        >>>         vals = series.all_values(copy=False)
        >>>         fit_params = params['fitted']
        >>>         unit_scale = (vals - fit_params['position'])/fit_params['scale']
        >>>         fix_params = params['fixed']
        >>>         rescaled = fix_params['_scale'] * unit_scale + fix_params['_position']
        >>>         return series.from_values(rescaled)
        >>>
        >>>     @staticmethod
        >>>     def ts_fit(series, params):
        >>>         vals = series.all_values(copy=False)
        >>>         scale = vals.max() - vals.min()
        >>>         position = vals[0]
        >>>         return {'scale': scale, 'position': position}
        >>>
        >>> series = linear_timeseries(length=5, start_value=1, end_value=5)
        >>> print(series)
        <TimeSeries (DataArray) (time: 5, component: 1, sample: 1)>
        array([[[1.]],

            [[2.]],

            [[3.]],

            [[4.]],

            [[5.]]])
        Coordinates:
        * time       (time) datetime64[ns] 2000-01-01 2000-01-02 ... 2000-01-05
        * component  (component) object 'linear'
        Dimensions without coordinates: sample
        Attributes:
            static_covariates:  None
            hierarchy:          None
        >>> series = SimpleRangeScaler(scale=2, position=-1).fit_transform(series)
        >>> print(series)
        <TimeSeries (DataArray) (time: 5, component: 1, sample: 1)>
        array([[[-1. ]],

            [[-0.5]],

            [[ 0. ]],

            [[ 0.5]],

            [[ 1. ]]])
        Coordinates:
        * time       (time) int64 0 1 2 3 4
        * component  (component) <U1 '0'
        Dimensions without coordinates: sample
        Attributes:
            static_covariates:  None
            hierarchy:          None
        """
        super().__init__(
            name=name,
            n_jobs=n_jobs,
            verbose=verbose,
            parallel_params=parallel_params,
            mask_components=mask_components,
        )

        self._fit_called = False
        self._fitted_params = None  # stores the fitted parameters/objects
        self._global_fit = global_fit

    @staticmethod
    @abstractmethod
    def ts_fit(
        series: Union[TimeSeries, Sequence[TimeSeries]],
        params: Mapping[str, Any],
        *args,
        **kwargs,
    ):
        """The function that will be applied to each series when :func:`fit` is called.

        If the `global_fit` attribute is set to `False`, then `ts_fit` should accept a `TimeSeries` as a first
        argument and return a set of parameters that are fitted to this individual `TimeSeries`. Conversely, if the
        `global_fit` attribute is set to `True`, then `ts_fit` should accept a `Sequence[TimeSeries]` and
        return a set of parameters that are fitted to *all* of the provided `TimeSeries`. All these parameters will
        be stored in ``self._fitted_params``, which can be later used during the transformation step.

        Regardless of whether the `global_fit` attribute is set to `True` or `False`, `ts_fit` should also accept
        a dictionary of fixed parameter values as a second argument (i.e. `params['fixed'] contains the fixed
        parameters of the data transformer).

        Any additional positional and/or keyword arguments passed to the `fit` method will be passed as
        positional/keyword arguments to `ts_fit`.

        This method is not implemented in the base class and must be implemented in the deriving classes.

        If more parameters are added as input in the derived classes, :func:`_fit_iterator()`
        should be redefined accordingly, to yield the necessary arguments to this function (See
        :func:`_fit_iterator()` for further details)

        Parameters
        ----------
        series (Union[TimeSeries, Sequence[TimeSeries]])
            `TimeSeries` against which the scaler will be fit.

        Notes
        -----
        This method is designed to be a static method instead of instance methods to allow an efficient
        parallelisation also when the scaler instance is storing a non-negligible amount of data. Using instance
        methods would imply copying the instance's data through multiple processes, which can easily introduce a
        bottleneck and nullify parallelisation benefits.
        """
        pass

    def fit(
        self,
        series: Union[TimeSeries, Sequence[TimeSeries]],
        *args,
        component_mask: Optional[np.array] = None,
        **kwargs,
    ) -> "FittableDataTransformer":
        """Fits transformer to a (sequence of) `TimeSeries` by calling the user-implemented `ts_fit` method.

        The fitted parameters returned by `ts_fit` are stored in the ``self._fitted_params`` attribute.
        If a `Sequence[TimeSeries]` is passed as the `series` data, then one of two outcomes will occur:
            1. If the `global_fit` attribute was set to `False`, then a different set of parameters will be
            individually fitted to each `TimeSeries` in the `Sequence`. In this case, this function automatically
            parallelises this fitting process over all of the multiple `TimeSeries` that have been passed.
            2. If the `global_fit` attribute was set to `True`, then all of the `TimeSeries` objects will be used
            fit a single set of parameters.

        Parameters
        ----------
        series
            (sequence of) series to fit the transformer on.
        args
            Additional positional arguments for the :func:`ts_fit` method
        component_mask : Optional[np.ndarray] = None
            Optionally, a 1-D boolean np.ndarray of length ``series.n_components`` that specifies which
            components of the underlying `series` the transform should be fitted to.
        kwargs
            Additional keyword arguments for the :func:`ts_fit` method

        Returns
        -------
        FittableDataTransformer
            Fitted transformer.
        """
        self._fit_called = True

        desc = f"Fitting ({self._name})"

        if isinstance(series, TimeSeries):
            data = [series]
        else:
            data = series

        if self._mask_components:
            data = [
                self.apply_component_mask(ts, component_mask, return_ts=True)
                for ts in data
            ]
        else:
            kwargs["component_mask"] = component_mask

        params_iterator = self._get_params(n_timeseries=len(data), calling_fit=True)
        fit_iterator = (
            zip(data, params_iterator)
            if not self._global_fit
            else zip([data], params_iterator)
        )
        n_jobs = len(data) if not self._global_fit else 1
        input_iterator = _build_tqdm_iterator(
            fit_iterator, verbose=self._verbose, desc=desc, total=n_jobs
        )

        self._fitted_params = _parallel_apply(
            input_iterator, self.__class__.ts_fit, self._n_jobs, args, kwargs
        )

        return self

    def fit_transform(
        self,
        series: Union[TimeSeries, Sequence[TimeSeries]],
        *args,
        component_mask: Optional[np.array] = None,
        **kwargs,
    ) -> Union[TimeSeries, List[TimeSeries]]:
        """Fit the transformer to the (sequence of) series and return the transformed input.

        Parameters
        ----------
        series
            the (sequence of) series to transform.
        args
            Additional positional arguments passed to the :func:`ts_transform` and :func:`ts_fit` methods.
        component_mask : Optional[np.ndarray] = None
            Optionally, a 1-D boolean np.ndarray of length ``series.n_components`` that specifies which
            components of the underlying `series` the transform should be fitted and applied to.
        kwargs
            Additional keyword arguments passed to the :func:`ts_transform` and :func:`ts_fit` methods.

        Returns
        -------
        Union[TimeSeries, Sequence[TimeSeries]]
            Transformed data.
        """
        return self.fit(
            series, *args, component_mask=component_mask, **kwargs
        ).transform(series, *args, component_mask=component_mask, **kwargs)

    def _get_params(
        self, n_timeseries: int, calling_fit: bool = False
    ) -> Generator[Mapping[str, Any], None, None]:
        """
        Overrides `_get_params` of `BaseDataTransformer`. Creates generator of dictionaries containing
        both the fixed parameter values (i.e. attributes defined in the child-most class), as well as
        the fitted parameter values (only if `calling_fit = False`). Those fixed parameters
        specified by `parallel_params` are given different values over each of the parallel jobs;
        every fitted parameter is given a different value for each parallel job (since `self.fit`
        returns a `Sequence` containing one fitted parameter value of each parallel job). Called by
        `transform` and `inverse_transform`.
        """
        # Call `_check_fixed_params` of `BaseDataTransformer`:
        self._check_fixed_params(n_timeseries)
        fitted_params = self._get_fitted_params(n_timeseries, calling_fit)

        def params_generator(
            n_jobs, fixed_params, fitted_params, parallel_params, global_fit
        ):
            fixed_params_copy = fixed_params.copy()
            for i in range(n_jobs):
                for key in parallel_params:
                    fixed_params_copy[key] = fixed_params[key][i]
                params = {}
                if fixed_params_copy:
                    params["fixed"] = fixed_params_copy
                if fitted_params:
                    params["fitted"] = (
                        fitted_params[0] if global_fit else fitted_params[i]
                    )
                if not params:
                    params = None
                yield params

        n_jobs = n_timeseries if not (calling_fit and self._global_fit) else 1

        return params_generator(
            n_jobs,
            self._fixed_params,
            fitted_params,
            self._parallel_params,
            self._global_fit,
        )

    def _get_fitted_params(self, n_timeseries: int, calling_fit: bool) -> Sequence[Any]:
        """
        Returns `self._fitted_params` if `calling_fit = False`, otherwise returns an empty
        tuple. If `calling_fit = False`, also checks that `self._fitted_params`, which is a
        sequence of values, contains exactly `n_timeseries` values; if not, a `ValueError` is thrown.
        """
        if not calling_fit:
            raise_if_not(
                self._fit_called,
                ("Must call `fit` before calling `transform`/`inverse_transform`."),
            )
            fitted_params = self._fitted_params
        else:
            fitted_params = tuple()
        if not self._global_fit and fitted_params:
            raise_if(
                n_timeseries > len(fitted_params),
                (
                    f"{n_timeseries} TimeSeries were provided "
                    f"but only {len(fitted_params)} TimeSeries "
                    f"were specified upon training {self.name}."
                ),
            )
        return fitted_params



import shutil
import tempfile

import pandas as pd

from darts import TimeSeries
from darts.logging import get_logger
from darts.tests.base_test_class import DartsBaseTestClass
from darts.utils import timeseries_generation as tg

logger = get_logger(__name__)

try:
    import torch.nn as nn

    from darts.models.components.transformer import (
        CustomFeedForwardDecoderLayer,
        CustomFeedForwardEncoderLayer,
    )
    from darts.models.forecasting.transformer_model import (
        TransformerModel,
        _TransformerModule,
    )

    TORCH_AVAILABLE = True
except ImportError:
    logger.warning("Torch not available. Transformer tests will be skipped.")
    TORCH_AVAILABLE = False


if TORCH_AVAILABLE:

    class TransformerModelTestCase(DartsBaseTestClass):
        __test__ = True
        times = pd.date_range("20130101", "20130410")
        pd_series = pd.Series(range(100), index=times)
        series: TimeSeries = TimeSeries.from_series(pd_series)
        series_multivariate = series.stack(series * 2)
        module = _TransformerModule(
            input_size=1,
            input_chunk_length=1,
            output_chunk_length=1,
            output_size=1,
            nr_params=1,
            d_model=512,
            nhead=8,
            num_encoder_layers=6,
            num_decoder_layers=6,
            dim_feedforward=2048,
            dropout=0.1,
            activation="relu",
            norm_type=None,
            custom_encoder=None,
            custom_decoder=None,
        )

        def setUp(self):
            self.temp_work_dir = tempfile.mkdtemp(prefix="darts")

        def tearDown(self):
            shutil.rmtree(self.temp_work_dir)

        def test_fit(self):
            # Test fit-save-load cycle
            model2 = TransformerModel(
                input_chunk_length=1,
                output_chunk_length=1,
                n_epochs=2,
                model_name="unittest-model-transformer",
                work_dir=self.temp_work_dir,
                save_checkpoints=True,
                force_reset=True,
            )
            model2.fit(self.series)
            model_loaded = model2.load_from_checkpoint(
                model_name="unittest-model-transformer",
                work_dir=self.temp_work_dir,
                best=False,
            )
            pred1 = model2.predict(n=6)
            pred2 = model_loaded.predict(n=6)

            # Two models with the same parameters should deterministically yield the same output
            self.assertEqual(sum(pred1.values() - pred2.values()), 0.0)

            # Another random model should not
            model3 = TransformerModel(
                input_chunk_length=1, output_chunk_length=1, n_epochs=1
            )
            model3.fit(self.series)
            pred3 = model3.predict(n=6)
            self.assertNotEqual(sum(pred1.values() - pred3.values()), 0.0)

            # test short predict
            pred4 = model3.predict(n=1)
            self.assertEqual(len(pred4), 1)

            # test validation series input
            model3.fit(self.series[:60], val_series=self.series[60:])
            pred4 = model3.predict(n=6)
            self.assertEqual(len(pred4), 6)

        def helper_test_pred_length(self, pytorch_model, series):
            model = pytorch_model(
                input_chunk_length=1, output_chunk_length=3, n_epochs=1
            )
            model.fit(series)
            pred = model.predict(7)
            self.assertEqual(len(pred), 7)
            pred = model.predict(2)
            self.assertEqual(len(pred), 2)
            self.assertEqual(pred.width, 1)
            pred = model.predict(4)
            self.assertEqual(len(pred), 4)
            self.assertEqual(pred.width, 1)

        def test_pred_length(self):
            series = tg.linear_timeseries(length=100)
            self.helper_test_pred_length(TransformerModel, series)

        def test_activations(self):
            with self.assertRaises(ValueError):
                model1 = TransformerModel(
                    input_chunk_length=1, output_chunk_length=1, activation="invalid"
                )
                model1.fit(self.series, epochs=1)

            # internal activation function uses PyTorch TransformerEncoderLayer
            model2 = TransformerModel(
                input_chunk_length=1, output_chunk_length=1, activation="gelu"
            )
            model2.fit(self.series, epochs=1)
            assert isinstance(
                model2.model.transformer.encoder.layers[0], nn.TransformerEncoderLayer
            )
            assert isinstance(
                model2.model.transformer.decoder.layers[0], nn.TransformerDecoderLayer
            )

            # glue variant FFN uses our custom _FeedForwardEncoderLayer
            model3 = TransformerModel(
                input_chunk_length=1, output_chunk_length=1, activation="SwiGLU"
            )
            model3.fit(self.series, epochs=1)
            assert isinstance(
                model3.model.transformer.encoder.layers[0],
                CustomFeedForwardEncoderLayer,
            )
            assert isinstance(
                model3.model.transformer.decoder.layers[0],
                CustomFeedForwardDecoderLayer,
            )

        def test_layer_norm(self):
            base_model = TransformerModel

            # default norm_type is None
            model0 = base_model(input_chunk_length=1, output_chunk_length=1)
            y0 = model0.fit(self.series, epochs=1)

            model1 = base_model(
                input_chunk_length=1, output_chunk_length=1, norm_type="RMSNorm"
            )
            y1 = model1.fit(self.series, epochs=1)

            model2 = base_model(
                input_chunk_length=1, output_chunk_length=1, norm_type=nn.LayerNorm
            )
            y2 = model2.fit(self.series, epochs=1)

            model3 = base_model(
                input_chunk_length=1,
                output_chunk_length=1,
                activation="gelu",
                norm_type="RMSNorm",
            )
            y3 = model3.fit(self.series, epochs=1)

            assert y0 != y1
            assert y0 != y2
            assert y0 != y3
            assert y1 != y3

            with self.assertRaises(AttributeError):
                model4 = base_model(
                    input_chunk_length=1, output_chunk_length=1, norm_type="invalid"
                )
                model4.fit(self.series, epochs=1)



"""
Data Transformer Base Class
---------------------------
"""

from abc import ABC, abstractmethod
from typing import Any, Generator, List, Mapping, Optional, Sequence, Union

import numpy as np
import xarray as xr

from darts import TimeSeries
from darts.logging import get_logger, raise_if, raise_if_not
from darts.utils import _build_tqdm_iterator, _parallel_apply

logger = get_logger(__name__)


class BaseDataTransformer(ABC):
    def __init__(
        self,
        name: str = "BaseDataTransformer",
        n_jobs: int = 1,
        verbose: bool = False,
        parallel_params: Union[bool, Sequence[str]] = False,
        mask_components: bool = True,
    ):
        """Abstract class for data transformers.

        All the deriving classes have to implement the static method :func:`ts_transform`; this implemented
        method can then be applied to ``TimeSeries`` or ``Sequence[TimeSeries]`` inputs by calling the
        :func:`transform()` method. Internally, :func:`transform()` parallelizes func:`ts_transform` over all
        of the ``TimeSeries`` inputs passed to it. See the func:`ts_transform` method documentation
        for further details on how to implement this method in a user-defined class.

        Data transformers requiring to be fit first before calling :func:`transform` should derive
        from :class:`.FittableDataTransformer` instead. Data transformers that are invertible should derive
        from :class:`.InvertibleDataTransformer` instead. Transformers which are both fittable and invertible
        should inherit from both :class:`.FittableDataTransformer` and :class:`.InvertibleDataTransformer`.

        All Data Transformers can store *fixed parameters* that are automatically passed to func:`ts_transform`;
        the fixed parameters of a data transformer object are taken to be all those attributes defined
        in the `__init__` method of the child-most class *before* `super().__init__` is called. The fixed parameter
        values can then be accessed within the func:`ts_transform` method through the `params` dictionary
        argument. More specifically, `params['fixed']` stores a dictionary with all of the fixed parameter
        values, where the keys are simply the attribute names of each fixed parameter (e.g. the `self._my_param`
        fixed parameter attribute is accessed through `params['fixed']['_my_param']`).

        Data Transformers which inherit from :class:`.FittableDataTransformer` can also store fitted parameters
        alongside fixed parameters; please refer to the :class:`.FittableDataTransformer` documentation for
        further details.

        Note: the :func:`ts_transform` method is designed to be a static method instead of a instance method to allow an
        efficient parallelisation also when the scaler instance is storing a non-negligible amount of data. Using
        an instance method would imply copying the instance's data through multiple processes, which can easily
        introduce a bottleneck and nullify parallelisation benefits.

        Parameters
        ----------
        name
            The data transformer's name
        n_jobs
            The number of jobs to run in parallel. Parallel jobs are created only when a ``Sequence[TimeSeries]``
            is passed as input to a method, parallelising operations regarding different TimeSeries.
            Defaults to `1` (sequential). Setting the parameter to `-1` means using all the available processors.
            Note: for a small amount of data, the parallelisation overhead could end up increasing the total
            required amount of time.
        verbose
            Optionally, whether to print operations progress
        parallel_params
            Optionally, specifies which fixed parameters (i.e. the attributes initialized in the child-most
            class's `__init__`) take on different values for different parallel jobs. Fixed parameters specified
            by `parallel_params` are assumed to be a `Sequence` of values that should be used for that parameter
            in each parallel job; the length of this `Sequence` should equal the number of parallel jobs. If
            `parallel_params=True`, every fixed parameter will take on a different value for each
            parallel job. If `parallel_params=False`, every fixed parameter will take on the same value for
            each parallel job. If `parallel_params` is a `Sequence` of fixed attribute names, only those
            attribute names specified will take on different values between different parallel jobs.
        mask_components
            Optionally, whether or not to automatically apply any provided `component_mask`s to the
            `TimeSeries` inputs passed to `transform`, `fit`, `inverse_transform`, or `fit_transform`.
            If `True`, any specified `component_mask` will be applied to each input timeseries
            before passing them to the called method; the masked components will also be automatically
            'unmasked' in the returned `TimeSeries`. If `False`, then `component_mask` (if provided) will
            be passed as a keyword argument, but won't automatically be applied to the input timeseries.
            See `apply_component_mask` for further details.

        Example
        --------
        >>> from darts.dataprocessing.transformers import BaseDataTransformer
        >>> from darts.utils.timeseries_generation import linear_timeseries
        >>>
        >>> class SimpleTransform(BaseDataTransformer):
        >>>
        >>>         def __init__(self, a):
        >>>             self._a = a
        >>>             super().__init__()
        >>>
        >>>         @staticmethod
        >>>         def ts_transform(series, params, **kwargs):
        >>>             a = params['fixed']['_a']
        >>>             b = kwargs.pop('b')
        >>>             return a*series + b
        >>>
        >>> series = linear_timeseries(length=5)
        >>> print(series)
        <TimeSeries (DataArray) (time: 5, component: 1, sample: 1)>
        array([[[0.  ]],

            [[0.25]],

            [[0.5 ]],

            [[0.75]],

            [[1.  ]]])
        Coordinates:
        * time       (time) datetime64[ns] 2000-01-01 2000-01-02 ... 2000-01-05
        * component  (component) object 'linear'
        Dimensions without coordinates: sample
        Attributes:
            static_covariates:  None
            hierarchy:          None
        >>> series = SimpleTransform(a=2).transform(series, b=3)
        >>> print(series)
        <TimeSeries (DataArray) (time: 5, component: 1, sample: 1)>
        array([[[3. ]],

            [[3.5]],

            [[4. ]],

            [[4.5]],

            [[5. ]]])
        Coordinates:
        * time       (time) datetime64[ns] 2000-01-01 2000-01-02 ... 2000-01-05
        * component  (component) object 'linear'
        Dimensions without coordinates: sample
        Attributes:
            static_covariates:  None
            hierarchy:          None
        """
        # Assume `super().__init__` called at *very end* of
        # child-most class's `__init__`, so `vars(self)` contains
        # only attributes of this child-most class:
        self._fixed_params = vars(self).copy()
        # If `parallel_params = True`, but not a list of key values:
        if parallel_params and not isinstance(parallel_params, Sequence):
            parallel_params = self._fixed_params.keys()
        # If `parallel_params = False`:
        elif not parallel_params:
            parallel_params = tuple()
        self._parallel_params = parallel_params
        self._mask_components = mask_components
        self._name = name
        self._verbose = verbose
        self._n_jobs = n_jobs

    def set_verbose(self, value: bool):
        """Set the verbosity status.

        `True` for enabling the detailed report about scaler's operation progress,
        `False` for no additional information.

        Parameters
        ----------
        value
            New verbosity status
        """
        raise_if_not(isinstance(value, bool), "Verbosity status must be a boolean.")

        self._verbose = value

    def set_n_jobs(self, value: int):
        """Set the number of processors to be used by the transformer while processing multiple ``TimeSeries``.

        Parameters
        ----------
        value
            New n_jobs value.  Set to `-1` for using all the available cores.
        """

        raise_if_not(isinstance(value, int), "n_jobs must be an integer")
        self._n_jobs = value

    @staticmethod
    @abstractmethod
    def ts_transform(series: TimeSeries, params: Mapping[str, Any]) -> TimeSeries:
        """The function that will be applied to each series when :func:`transform()` is called.

        This method is not implemented in the base class and must be implemented in the deriving classes.

        The function must take as first argument a ``TimeSeries`` object and, as a second argument, a
        dictionary containing the fixed and/or fitted parameters of the transformation; this function
        should then return a transformed ``TimeSeries`` object.

        The `params` dictionary *can* contain up to two keys:
            1. `params['fixed']` stores the fixed parameters of the transformation (i.e. attributed
            defined in the `__init__` method of the child-most class *before* `super().__init__` is called);
            `params['fixed']` is a dictionary itself, whose keys are the names of the fixed parameter
            attributes. For example, if `_my_fixed_param` is defined as an attribute in the child-most
            class, then this fixed parameter value can be accessed through `params['fixed']['_my_fixed_param']`.
            2. If the transform inherits from the :class:`.FittableDataTransformer` class, then `params['fitted']`
            will store the fitted parameters of the transformation; the fitted parameters are simply the output(s)
            returned by the `ts_fit` function, whatever those output(s) may be. See :class:`.FittableDataTransformer`
            for further details about fitted parameters.

        Any positional/keyword argument supplied to the `transform` method are passed as positional/keyword arguments
        to `ts_transform`; hence, `ts_transform` should also accept `*args` and/or `**kwargs` if positional/keyword
        arguments are passed to `transform`. Note that if the `mask_components` attribute of `BaseDataTransformer`
        is set to `False`, then the `component_mask` provided to `transform` will be passed as an additional keyword
        argument to `ts_transform`.

        The `BaseDataTransformer` class includes some helper methods which may prove useful when implementing a
        `ts_transform` function:
            1. The `apply_component_mask` and `unapply_component_mask` methods, which apply and 'unapply'
            `component_mask`s to a `TimeSeries` respectively; these methods are automatically called in `transform`
            if the `mask_component` attribute of `BaseDataTransformer` is set to `True`, but you may want to manually
            call them if you set `mask_components` to `False` and wish to manually specify how `component_mask`s are
            applied to a `TimeSeries`.
            2. The `stack_samples` method, which stacks all the samples in a `TimeSeries` along
            the component axis, so that the `TimeSeries` goes from shape `(n_timesteps, n_components, n_samples)` to
            shape `(n_timesteps, n_components * n_samples)`. This stacking is useful if a pointwise transform is being
            implemented (i.e. transforming the value at time `t` depends only on the value of the series at that
            time `t`). Once transformed, the stacked `TimeSeries` can be 'unstacked' using the `unstack_samples` method.

        Parameters
        ----------
        series
            series to be transformed.
        params
            Dictionary containing the parameters of the transformation function. Fixed parameters
            (i.e. attributes defined in the child-most class of the transformation prior to
            calling `super.__init__()`) are stored under the `'fixed'` key. If the transformation
            inherits from the `FittableDataTransformer` class, then the fitted parameters of the
            transformation (i.e. the values returned by `ts_fit`) are stored under the
            `'fitted'` key.
        args
            Any poisitional arguments provided in addition to `series` when
        kwargs
            Any additional keyword arguments provided to `transform`. Note that if the `mask_component`
            attribute of `BaseDataTransformer` is set to `False`, then `component_mask` will
            be passed as a keyword argument.

        Notes
        -----
        This method is designed to be a static method instead of instance method to allow an efficient
        parallelisation also when the scaler instance is storing a non-negligible amount of data. Using instance
        methods would imply copying the instance's data through multiple processes, which can easily introduce a
        bottleneck and nullify parallelisation benefits.
        """
        pass

    def transform(
        self,
        series: Union[TimeSeries, Sequence[TimeSeries]],
        *args,
        component_mask: Optional[np.array] = None,
        **kwargs,
    ) -> Union[TimeSeries, List[TimeSeries]]:
        """Transforms a (sequence of) of series by calling the user-implemeneted `ts_transform` method.

        In case a ``Sequence[TimeSeries]`` is passed as input data, this function takes care of
        parallelising the transformation of multiple series in the sequence at the same time. Additionally,
        if the `mask_components` attribute was set to `True` when instantiating `BaseDataTransformer`,
        then any provided `component_mask`s will be automatically applied to each input `TimeSeries`;
        please refer to 'Notes' for further details on component masking.

        Any additionally specified `*args` and `**kwargs` are automatically passed to `ts_transform`.

        Parameters
        ----------
        series
            (sequence of) series to be transformed.
        args
            Additional positional arguments for each :func:`ts_transform()` method call
        component_mask : Optional[np.ndarray] = None
            Optionally, a 1-D boolean np.ndarray of length ``series.n_components`` that specifies which
            components of the underlying `series` the transform should consider. If the `mask_components`
            attribute was set to `True` when instantiating `BaseDataTransformer`, then the component mask
            will be automatically applied to each `TimeSeries` input. Otherwise, `component_mask` will be
            provided as an addition keyword argument to `ts_transform`. See 'Notes' for further details.
        kwargs
            Additional keyword arguments for each :func:`ts_transform()` method call

        Returns
        -------
        Union[TimeSeries, List[TimeSeries]]
            Transformed data.

        Notes
        -----
        If the `mask_components` attribute was set to `True` when instantiating `BaseDataTransformer`,
        then any provided `component_mask`s will be automatically applied to each `TimeSeries` input to
        transform; `component_mask`s are simply boolean arrays of shape `(series.n_components,)` that
        specify which components of each `series` should be transformed using `ts_transform` and which
        components should not. If `component_mask[i]` is `True`, then the `i`th component of each
        `series` will be transformed by `ts_transform`. Conversely, if `component_mask[i]` is `False`,
        the `i`th component will be removed from each `series` before being passed to `ts_transform`;
        after transforming this masked series, the untransformed `i`th component will be 'added back'
        to the output. Note that automatic `component_mask`ing can only be performed if the `ts_transform`
        does *not* change the number of timesteps in each series; if this were to happen, then the transformed
        and untransformed components are unable to be concatenated back together along the component axis.

        If `mask_components` was set to `False` when instantiating `BaseDataTransformer`, then any provided
        `component_masks` will be passed as a keyword argument `ts_transform`; the user can then manually specify
        how the `component_mask` should be applied to each series.
        """

        desc = f"Transform ({self._name})"

        # Take note of original input for unmasking purposes:
        if isinstance(series, TimeSeries):
            input_series = [series]
            data = [series]
        else:
            input_series = series
            data = series

        if self._mask_components:
            data = [
                self.apply_component_mask(ts, component_mask, return_ts=True)
                for ts in data
            ]
        else:
            kwargs["component_mask"] = component_mask

        input_iterator = _build_tqdm_iterator(
            zip(data, self._get_params(n_timeseries=len(data))),
            verbose=self._verbose,
            desc=desc,
            total=len(data),
        )

        transformed_data = _parallel_apply(
            input_iterator, self.__class__.ts_transform, self._n_jobs, args, kwargs
        )

        if self._mask_components:
            unmasked = []
            for ts, transformed_ts in zip(input_series, transformed_data):
                unmasked.append(
                    self.unapply_component_mask(ts, transformed_ts, component_mask)
                )
            transformed_data = unmasked

        return (
            transformed_data[0] if isinstance(series, TimeSeries) else transformed_data
        )

    def _get_params(
        self, n_timeseries: int
    ) -> Generator[Mapping[str, Any], None, None]:
        """
        Creates generator of dictionaries containing fixed parameter values
        (i.e. attributes defined in the child-most class). Those fixed parameters
        specified by `parallel_params` are given different values over each of the
        parallel jobs. Called by `transform` and `inverse_transform`,
        if `Transformer` does *not* inherit from `FittableTransformer`.
        """
        self._check_fixed_params(n_timeseries)

        def params_generator(n_timeseries, fixed_params, parallel_params):
            fixed_params_copy = fixed_params.copy()
            for i in range(n_timeseries):
                for key in parallel_params:
                    fixed_params_copy[key] = fixed_params[key][i]
                if fixed_params_copy:
                    params = {"fixed": fixed_params_copy}
                else:
                    params = None
                yield params
            return None

        return params_generator(n_timeseries, self._fixed_params, self._parallel_params)

    def _check_fixed_params(self, n_timeseries: int) -> None:
        """
        Raises `ValueError` if `self._parallel_params` specifies a `key` in
        `self._fixed_params` that should be distributed, but
        `len(self._fixed_params[key])` does not equal `n_timeseries`.
        """
        for key in self._parallel_params:
            raise_if(
                n_timeseries > len(self._fixed_params[key]),
                f"{n_timeseries} TimeSeries were provided "
                f"but only {len(self._fixed_params[key])} {key} values "
                f"were specified upon initialising {self.name}.",
            )
        return None

    @staticmethod
    def apply_component_mask(
        series: TimeSeries, component_mask: Optional[np.ndarray] = None, return_ts=False
    ) -> np.ndarray:
        """
        Extracts components specified by `component_mask` from `series`

        Parameters
        ----------
        series
            input TimeSeries to be fed into transformer.
        component_mask
            Optionally, np.ndarray boolean mask of shape (n_components, 1) specifying which components to
            extract from `series`. The `i`th component of `series` is kept only if `component_mask[i] = True`.
            If not specified, no masking is performed.
        return_ts
            Optionally, specifies that a `TimeSeries` should be returned, rather than an `np.ndarray`.

        Returns
        -------
        masked
            `TimeSeries` (if `return_ts = True`) or `np.ndarray` (if `return_ts = False`) with only those components
            specified by `component_mask` remaining.

        """
        if component_mask is None:
            masked = series.copy() if return_ts else series.all_values()
        else:
            raise_if_not(
                isinstance(component_mask, np.ndarray) and component_mask.dtype == bool,
                f"`component_mask` must be a boolean `np.ndarray`, not a {type(component_mask)}.",
                logger,
            )
            raise_if_not(
                series.width == len(component_mask),
                "mismatch between number of components in `series` and length of `component_mask`",
                logger,
            )
            masked = series.all_values(copy=False)[:, component_mask, :]
            if return_ts:
                # Remove masked components from coords:
                coords = dict(series._xa.coords)
                coords["component"] = coords["component"][component_mask]
                new_xa = xr.DataArray(
                    masked, dims=series._xa.dims, coords=coords, attrs=series._xa.attrs
                )
                masked = TimeSeries(new_xa)
        return masked

    @staticmethod
    def unapply_component_mask(
        series: TimeSeries,
        vals: Union[np.ndarray, TimeSeries],
        component_mask: Optional[np.ndarray] = None,
    ) -> Union[np.ndarray, TimeSeries]:
        """
        Adds back components previously removed by `component_mask` in `apply_component_mask` method.

        Parameters
        ----------
        series
            input TimeSeries that was fed into transformer.
        vals:
            `np.ndarray` or `TimeSeries` to 'unmask'
        component_mask
            Optionally, np.ndarray boolean mask of shape (n_components, 1) specifying which components were extracted
            from `series`. If given, insert `vals` back into the columns of the original array. If not specified,
            nothing is 'unmasked'.

        Returns
        -------
        unmasked
            `TimeSeries` (if `vals` is a `TimeSeries`) or `np.ndarray` (if `vals` is an `np.ndarray`) with those
            components previously removed by `component_mask` now 'added back'.
        """

        if component_mask is None:
            unmasked = vals
        else:
            raise_if_not(
                isinstance(component_mask, np.ndarray) and component_mask.dtype == bool,
                "If `component_mask` is given, must be a boolean np.ndarray`",
                logger,
            )
            raise_if_not(
                series.width == len(component_mask),
                "mismatch between number of components in `series` and length of `component_mask`",
                logger,
            )
            unmasked = series.all_values()
            if isinstance(vals, TimeSeries):
                unmasked[:, component_mask, :] = vals.all_values()
                # Remove timepoints not present in transformed data:
                unmasked = series.slice_intersect(vals).with_values(unmasked)
            else:
                unmasked[:, component_mask, :] = vals
        return unmasked

    @staticmethod
    def stack_samples(vals: Union[np.ndarray, TimeSeries]) -> np.ndarray:
        """
        Creates an array of shape `(n_timesteps * n_samples, n_components)` from
        either a `TimeSeries` or the `array_values` of a `TimeSeries`.

        Each column of the returned array corresponds to a component (dimension)
        of the series and is formed by concatenating all of the samples associated
        with that component together. More specifically, the `i`th column is formed
        by concatenating
        `[component_i_sample_1, component_i_sample_2, ..., component_i_sample_n]`.

        Stacking is useful when implementing a transformation that applies the exact same
        change to every timestep in the timeseries. In such cases, the samples of each
        component can be stacked together into a single column, and the transformation can then be
        applied to each column, thereby 'vectorising' the transformation over all samples of that component;
        the `unstack_samples` method can then be used to reshape the output. For transformations that depend
        on the `time_index` or the temporal ordering of the observations, stacking should *not* be employed.

        Parameters
        ----------
        vals
            `Timeseries` or `np.ndarray` of shape `(n_timesteps, n_components, n_samples)` to be 'stacked'.

        Returns
        -------
        stacked
            `np.ndarray` of shape `(n_timesteps * n_samples, n_components)`, where the `i`th column is formed
            by concatenating all of the samples of the `i`th component in `vals`.
        """
        if isinstance(vals, TimeSeries):
            vals = vals.all_values()
        shape = vals.shape
        new_shape = (shape[0] * shape[2], shape[1])
        stacked = np.swapaxes(vals, 1, 2).reshape(new_shape)
        return stacked

    @staticmethod
    def unstack_samples(
        vals: np.ndarray,
        n_timesteps: Optional[int] = None,
        n_samples: Optional[int] = None,
        series: Optional[TimeSeries] = None,
    ) -> np.ndarray:
        """
        Reshapes the 2D array returned by `stack_samples` back into an array of shape
        `(n_timesteps, n_components, n_samples)`; this 'undoes' the reshaping of
        `stack_samples`. Either `n_components`, `n_samples`, or `series` must be specified.

        Parameters
        ----------
        vals
            `np.ndarray` of shape `(n_timesteps * n_samples, n_components)` to be 'unstacked'.
        n_timesteps
            Optionally, the number of timesteps in the array originally passed to `stack_samples`.
            Does *not* need to be provided if `series` is specified.
        n_samples
            Optionally, the number of samples in the array originally passed to `stack_samples`.
            Does *not* need to be provided if `series` is specified.
        series
            Optionally, the `TimeSeries` object used to create `vals`; `n_samples` is inferred
            from this.

        Returns
        -------
        unstacked
            `np.ndarray` of shape `(n_timesteps, n_components, n_samples)`.

        """
        if series is not None:
            n_samples = series.n_samples
        else:
            raise_if(
                all(x is None for x in [n_timesteps, n_samples]),
                "Must specify either `n_timesteps`, `n_samples`, or `series`.",
            )
        n_components = vals.shape[-1]
        if n_timesteps is not None:
            reshaped_vals = vals.reshape(n_timesteps, -1, n_components)
        else:
            reshaped_vals = vals.reshape(-1, n_samples, n_components)
        unstacked = np.swapaxes(reshaped_vals, 1, 2)
        return unstacked

    @property
    def name(self):
        """Name of the data transformer."""
        return self._name

    def __str__(self):
        return self._name

    def __repr__(self):
        return self.__str__()



"""
N-Linear
------
"""

from typing import Optional, Tuple

import torch
import torch.nn as nn

from darts.logging import raise_if
from darts.models.forecasting.pl_forecasting_module import PLMixedCovariatesModule
from darts.models.forecasting.torch_forecasting_model import MixedCovariatesTorchModel


class _NLinearModule(PLMixedCovariatesModule):
    """
    NLinear module
    """

    def __init__(
        self,
        input_dim,
        output_dim,
        future_cov_dim,
        static_cov_dim,
        nr_params,
        shared_weights,
        const_init,
        normalize,
        **kwargs,
    ):
        """PyTorch module implementing the N-HiTS architecture.

        Parameters
        ----------
        input_dim
            The number of input components (target + optional covariate)
        output_dim
            Number of output components in the target
        future_cov_dim
            Number of components in the future covariates
        static_cov_dim
            Dimensionality of the static covariates
        nr_params
            The number of parameters of the likelihood (or 1 if no likelihood is used).
        shared_weights
            Whether to use shared weights for the components of the series.
            ** Ignores covariates when True. **
        normalize
            Whether to apply the "normalization" described in the paper.
        const_init
            Whether to initialize the weights to 1/in_len

        **kwargs
            all parameters required for :class:`darts.model.forecasting_models.PLForecastingModule` base class.

        Inputs
        ------
        x of shape `(batch_size, input_chunk_length)`
            Tensor containing the input sequence.

        Outputs
        -------
        y of shape `(batch_size, output_chunk_length, target_size/output_dim, nr_params)`
            Tensor containing the output of the NBEATS module.

        """
        super().__init__(**kwargs)
        self.input_dim = input_dim
        self.output_dim = output_dim
        self.future_cov_dim = future_cov_dim
        self.static_cov_dim = static_cov_dim
        self.nr_params = nr_params
        self.shared_weights = shared_weights
        self.const_init = const_init
        self.normalize = normalize

        def _create_linear_layer(in_dim, out_dim):
            layer = nn.Linear(in_dim, out_dim)
            if self.const_init:
                layer.weight = nn.Parameter(
                    (1.0 / in_dim) * torch.ones(layer.weight.shape)
                )
            return layer

        if self.shared_weights:
            layer_in_dim = self.input_chunk_length
            layer_out_dim = self.output_chunk_length * self.nr_params
        else:
            layer_in_dim = self.input_chunk_length * self.input_dim
            layer_out_dim = self.output_chunk_length * self.output_dim * self.nr_params

            # for static cov, we take the number of components of the target, times static cov dim
            layer_in_dim_static_cov = self.output_dim * self.static_cov_dim

        self.layer = _create_linear_layer(layer_in_dim, layer_out_dim)

        if self.future_cov_dim != 0:
            # future covariates layer acts on time steps independently
            self.linear_fut_cov = _create_linear_layer(
                self.future_cov_dim, self.output_dim * self.nr_params
            )
        if self.static_cov_dim != 0:
            self.linear_static_cov = _create_linear_layer(
                layer_in_dim_static_cov, layer_out_dim
            )

    def forward(
        self, x_in: Tuple[torch.Tensor, Optional[torch.Tensor], Optional[torch.Tensor]]
    ):
        """
        x_in
            comes as tuple `(x_past, x_future, x_static)` where `x_past` is the input/past chunk and `x_future`
            is the output/future chunk. Input dimensions are `(n_samples, n_time_steps, n_variables)`
        """
        x, x_future, x_static = x_in  # x: (batch, in_len, in_dim)
        batch, _, _ = x.shape

        if self.shared_weights:
            # discard covariates, to ensure that in_dim == out_dim
            x = x[:, :, : self.output_dim]
            x = x.permute(0, 2, 1)  # (batch, out_dim, in_len)

            if self.normalize:
                seq_last = x[:, :, -1:].detach()  # (batch, out_dim, 1)
                x = x - seq_last

            x = self.layer(x)  # (batch, out_dim, out_len * nr_params)

            if self.normalize:
                x = x + seq_last

            # extract nr_params
            x = x.view(batch, self.output_dim, self.output_chunk_length, self.nr_params)

            # permute back to (batch, out_len, out_dim, nr_params)
            x = x.permute(0, 2, 1, 3)
        else:
            if self.normalize:
                seq_last = x[:, -1:, :].detach()  # (batch, 1, in_dim)
                x = x - seq_last

            x = self.layer(x.view(batch, -1))  # (batch, out_len * out_dim * nr_params)
            x = x.view(
                batch, self.output_chunk_length, self.output_dim * self.nr_params
            )

            if self.normalize:
                x = x + seq_last  # Note: works only when nr_params == 1

            if self.future_cov_dim != 0:
                # x_future might be shorter than output_chunk_length when n < output_chunk_length
                # so we need to pad it with zeros at the end to match the output_chunk_length
                x_future = torch.nn.functional.pad(
                    input=x_future,
                    pad=(0, 0, 0, self.output_chunk_length - x_future.shape[1]),
                    mode="constant",
                    value=0,
                )

                fut_cov_output = self.linear_fut_cov(x_future)
                x = x + fut_cov_output.view(
                    batch, self.output_chunk_length, self.output_dim * self.nr_params
                )

            if self.static_cov_dim != 0:
                static_cov_output = self.linear_static_cov(x_static.reshape(batch, -1))
                x = x + static_cov_output.view(
                    batch, self.output_chunk_length, self.output_dim * self.nr_params
                )

            x = x.view(batch, self.output_chunk_length, self.output_dim, self.nr_params)

        return x


class NLinearModel(MixedCovariatesTorchModel):
    def __init__(
        self,
        input_chunk_length: int,
        output_chunk_length: int,
        shared_weights: bool = False,
        const_init: bool = True,
        normalize: bool = False,
        use_static_covariates: bool = True,
        **kwargs,
    ):
        """An implementation of the NLinear model, as presented in [1]_.

        This implementation is improved by allowing the optional use of past covariates,
        future covariates and static covariates, and by making the model optionally probabilistic.

        Parameters
        ----------
        input_chunk_length
            The length of the input sequence fed to the model.
        output_chunk_length
            The length of the forecast of the model.
        shared_weights
            Whether to use shared weights for all components of multivariate series.

            .. warning::
                When set to True, covariates will be ignored as a 1-to-1 mapping is
                required between input dimensions and output dimensions.
            ..

            Default: False.

        const_init
            Whether to initialize the weights to 1/in_len. If False, the default PyTorch
            initialization is used (default='True').
        normalize
            Whether to apply the simple "normalization" proposed in the paper, which consists
            in subtracting the last value of the input sequence from the input sequence. Default: False.

            .. note::
                This cannot be applied to probabilistic models.
            ..
        use_static_covariates
            Whether the model should use static covariate information in case the input `series` passed to ``fit()``
            contain static covariates. If ``True``, and static covariates are available at fitting time, will enforce
            that all target `series` have the same static covariate dimensionality in ``fit()`` and ``predict()`.
        **kwargs
            Optional arguments to initialize the pytorch_lightning.Module, pytorch_lightning.Trainer, and
            Darts' :class:`TorchForecastingModel`.

        loss_fn
            PyTorch loss function used for training.
            This parameter will be ignored for probabilistic models if the ``likelihood`` parameter is specified.
            Default: ``torch.nn.MSELoss()``.
        likelihood
            One of Darts' :meth:`Likelihood <darts.utils.likelihood_models.Likelihood>` models to be used for
            probabilistic forecasts. Default: ``None``.
        torch_metrics
            A torch metric or a ``MetricCollection`` used for evaluation. A full list of available metrics can be found
            at https://torchmetrics.readthedocs.io/en/latest/. Default: ``None``.
        optimizer_cls
            The PyTorch optimizer class to be used. Default: ``torch.optim.Adam``.
        optimizer_kwargs
            Optionally, some keyword arguments for the PyTorch optimizer (e.g., ``{'lr': 1e-3}``
            for specifying a learning rate). Otherwise, the default values of the selected ``optimizer_cls``
            will be used. Default: ``None``.
        lr_scheduler_cls
            Optionally, the PyTorch learning rate scheduler class to be used. Specifying ``None`` corresponds
            to using a constant learning rate. Default: ``None``.
        lr_scheduler_kwargs
            Optionally, some keyword arguments for the PyTorch learning rate scheduler. Default: ``None``.
        batch_size
            Number of time series (input and output sequences) used in each training pass. Default: ``32``.
        n_epochs
            Number of epochs over which to train the model. Default: ``100``.
        model_name
            Name of the model. Used for creating checkpoints and saving tensorboard data. If not specified,
            defaults to the following string ``"YYYY-mm-dd_HH_MM_SS_torch_model_run_PID"``, where the initial part
            of the name is formatted with the local date and time, while PID is the processed ID (preventing models
            spawned at the same time by different processes to share the same model_name). E.g.,
            ``"2021-06-14_09_53_32_torch_model_run_44607"``.
        work_dir
            Path of the working directory, where to save checkpoints and Tensorboard summaries.
            Default: current working directory.
        log_tensorboard
            If set, use Tensorboard to log the different parameters. The logs will be located in:
            ``"{work_dir}/darts_logs/{model_name}/logs/"``. Default: ``False``.
        nr_epochs_val_period
            Number of epochs to wait before evaluating the validation loss (if a validation
            ``TimeSeries`` is passed to the :func:`fit()` method). Default: ``1``.
        force_reset
            If set to ``True``, any previously-existing model with the same name will be reset (all checkpoints will
            be discarded). Default: ``False``.
        save_checkpoints
            Whether or not to automatically save the untrained model and checkpoints from training.
            To load the model from checkpoint, call :func:`MyModelClass.load_from_checkpoint()`, where
            :class:`MyModelClass` is the :class:`TorchForecastingModel` class that was used (such as :class:`TFTModel`,
            :class:`NBEATSModel`, etc.). If set to ``False``, the model can still be manually saved using
            :func:`save()` and loaded using :func:`load()`. Default: ``False``.
        add_encoders
            A large number of past and future covariates can be automatically generated with `add_encoders`.
            This can be done by adding multiple pre-defined index encoders and/or custom user-made functions that
            will be used as index encoders. Additionally, a transformer such as Darts' :class:`Scaler` can be added to
            transform the generated covariates. This happens all under one hood and only needs to be specified at
            model creation.
            Read :meth:`SequentialEncoder <darts.dataprocessing.encoders.SequentialEncoder>` to find out more about
            ``add_encoders``. Default: ``None``. An example showing some of ``add_encoders`` features:

            .. highlight:: python
            .. code-block:: python

                add_encoders={
                    'cyclic': {'future': ['month']},
                    'datetime_attribute': {'future': ['hour', 'dayofweek']},
                    'position': {'past': ['relative'], 'future': ['relative']},
                    'custom': {'past': [lambda idx: (idx.year - 1950) / 50]},
                    'transformer': Scaler()
                }
            ..
        random_state
            Control the randomness of the weights initialization. Check this
            `link <https://scikit-learn.org/stable/glossary.html#term-random_state>`_ for more details.
            Default: ``None``.
        pl_trainer_kwargs
            By default :class:`TorchForecastingModel` creates a PyTorch Lightning Trainer with several useful presets
            that performs the training, validation and prediction processes. These presets include automatic
            checkpointing, tensorboard logging, setting the torch device and more.
            With ``pl_trainer_kwargs`` you can add additional kwargs to instantiate the PyTorch Lightning trainer
            object. Check the `PL Trainer documentation
            <https://pytorch-lightning.readthedocs.io/en/stable/common/trainer.html>`_ for more information about the
            supported kwargs. Default: ``None``.
            Running on GPU(s) is also possible using ``pl_trainer_kwargs`` by specifying keys ``"accelerator",
            "devices", and "auto_select_gpus"``. Some examples for setting the devices inside the ``pl_trainer_kwargs``
            dict:

                - ``{"accelerator": "cpu"}`` for CPU,
                - ``{"accelerator": "gpu", "devices": [i]}`` to use only GPU ``i`` (``i`` must be an integer),
                - ``{"accelerator": "gpu", "devices": -1, "auto_select_gpus": True}`` to use all available GPUS.

                For more info, see here:
                https://pytorch-lightning.readthedocs.io/en/stable/common/trainer.html#trainer-flags , and
                https://pytorch-lightning.readthedocs.io/en/stable/accelerators/gpu_basic.html#train-on-multiple-gpus
            With parameter ``"callbacks"`` you can add custom or PyTorch-Lightning built-in callbacks to Darts'
            :class:`TorchForecastingModel`. Below is an example for adding EarlyStopping to the training process.
            The model will stop training early if the validation loss `val_loss` does not improve beyond
            specifications. For more information on callbacks, visit:
            `PyTorch Lightning Callbacks
            <https://pytorch-lightning.readthedocs.io/en/stable/extensions/callbacks.html>`_

            .. highlight:: python
            .. code-block:: python

                from pytorch_lightning.callbacks.early_stopping import EarlyStopping

                # stop training when validation loss does not decrease more than 0.05 (`min_delta`) over
                # a period of 5 epochs (`patience`)
                my_stopper = EarlyStopping(
                    monitor="val_loss",
                    patience=5,
                    min_delta=0.05,
                    mode='min',
                )

                pl_trainer_kwargs={"callbacks": [my_stopper]}
            ..

            Note that you can also use a custom PyTorch Lightning Trainer for training and prediction with optional
            parameter ``trainer`` in :func:`fit()` and :func:`predict()`.
        show_warnings
            whether to show warnings raised from PyTorch Lightning. Useful to detect potential issues of
            your forecasting use case. Default: ``False``.

        References
        ----------
        .. [1] Zeng, A., Chen, M., Zhang, L., & Xu, Q. (2022).
               Are Transformers Effective for Time Series Forecasting?. arXiv preprint arXiv:2205.13504.
        """
        super().__init__(**self._extract_torch_model_params(**self.model_params))

        # extract pytorch lightning module kwargs
        self.pl_module_params = self._extract_pl_module_params(**self.model_params)

        self.shared_weights = shared_weights
        self.const_init = const_init
        self.normalize = normalize
        self._considers_static_covariates = use_static_covariates

        raise_if(
            "likelihood" in self.model_params
            and self.model_params["likelihood"] is not None
            and self.normalize,
            "normalize = True cannot be used with probabilistic NLinearModel",
        )

    def _create_model(self, train_sample: Tuple[torch.Tensor]) -> torch.nn.Module:
        # samples are made of
        # (past_target, past_covariates, historic_future_covariates,
        #  future_covariates, static_covariates, future_target)

        raise_if(
            self.shared_weights
            and (train_sample[1] is not None or train_sample[2] is not None),
            "Covariates have been provided, but the model has been built with shared_weights=True."
            + "Please set shared_weights=False to use covariates.",
        )

        input_dim = train_sample[0].shape[1] + sum(
            # add past covariates dim and historic future covariates dim, if present
            train_sample[i].shape[1] if train_sample[i] is not None else 0
            for i in (1, 2)
        )
        future_cov_dim = train_sample[3].shape[1] if train_sample[3] is not None else 0

        # dimension is (component, static_dim), we extract static_dim
        static_cov_dim = train_sample[4].shape[1] if train_sample[4] is not None else 0

        output_dim = train_sample[-1].shape[1]

        nr_params = 1 if self.likelihood is None else self.likelihood.num_parameters

        return _NLinearModule(
            input_dim=input_dim,
            output_dim=output_dim,
            future_cov_dim=future_cov_dim,
            static_cov_dim=static_cov_dim,
            nr_params=nr_params,
            shared_weights=self.shared_weights,
            const_init=self.const_init,
            normalize=self.normalize,
            **self.pl_module_params,
        )

    @property
    def supports_static_covariates(self) -> bool:
        return True



"""
OR Aggregator
-------------

Aggregator that identifies a time step as anomalous if any of the components
is flagged as anomalous (logical OR).
"""


from typing import Sequence

from darts import TimeSeries
from darts.ad.aggregators.aggregators import NonFittableAggregator


class OrAggregator(NonFittableAggregator):
    def __init__(self) -> None:
        super().__init__()

    def __str__(self):
        return "OrAggregator"

    def _predict_core(self, series: Sequence[TimeSeries]) -> Sequence[TimeSeries]:
        return [s.sum(axis=1).map(lambda x: (x > 0).astype(s.dtype)) for s in series]



import random
import unittest
from itertools import product

import numpy as np
import pandas as pd
import pytest

from darts import TimeSeries
from darts.datasets import AirPassengersDataset, MonthlyMilkDataset
from darts.logging import get_logger
from darts.metrics import mape, r2_score
from darts.models import (
    ARIMA,
    FFT,
    ExponentialSmoothing,
    NaiveDrift,
    NaiveSeasonal,
    Theta,
)
from darts.tests.base_test_class import DartsBaseTestClass
from darts.utils.timeseries_generation import gaussian_timeseries as gt
from darts.utils.timeseries_generation import linear_timeseries as lt
from darts.utils.timeseries_generation import random_walk_timeseries as rt
from darts.utils.timeseries_generation import sine_timeseries as st

logger = get_logger(__name__)


try:
    from darts.models import (
        BlockRNNModel,
        LinearRegressionModel,
        RandomForest,
        TCNModel,
    )

    TORCH_AVAILABLE = True
except ImportError:
    logger.warning(
        "Torch models are not installed - will not be tested for backtesting"
    )
    TORCH_AVAILABLE = False


def get_dummy_series(
    ts_length: int, lt_end_value: int = 10, st_value_offset: int = 10
) -> TimeSeries:
    return (
        lt(length=ts_length, end_value=lt_end_value)
        + st(length=ts_length, value_y_offset=st_value_offset)
        + rt(length=ts_length)
    )


def compare_best_against_random(model_class, params, series, stride=1):

    # instantiate best model in expanding window mode
    np.random.seed(1)
    best_model_1, _, _ = model_class.gridsearch(
        params,
        series,
        forecast_horizon=10,
        stride=stride,
        metric=mape,
        start=series.time_index[-21],
    )

    # instantiate best model in split mode
    train, val = series.split_before(series.time_index[-10])
    best_model_2, _, _ = model_class.gridsearch(
        params, train, val_series=val, metric=mape
    )

    # instantiate model with random parameters from 'params'
    random.seed(1)
    random_param_choice = {}
    for key in params.keys():
        random_param_choice[key] = random.choice(params[key])
    random_model = model_class(**random_param_choice)

    # perform backtest forecasting on both models
    best_score_1 = best_model_1.backtest(
        series, start=series.time_index[-21], forecast_horizon=10
    )
    random_score_1 = random_model.backtest(
        series, start=series.time_index[-21], forecast_horizon=10
    )

    # perform train/val evaluation on both models
    best_model_2.fit(train)
    best_score_2 = mape(best_model_2.predict(len(val)), series)
    random_model = model_class(**random_param_choice)
    random_model.fit(train)
    random_score_2 = mape(random_model.predict(len(val)), series)

    # check whether best models are at least as good as random models
    expanding_window_ok = best_score_1 <= random_score_1
    split_ok = best_score_2 <= random_score_2

    return expanding_window_ok and split_ok


class BacktestingTestCase(DartsBaseTestClass):
    def test_backtest_forecasting(self):
        linear_series = lt(length=50)
        linear_series_int = TimeSeries.from_values(linear_series.values())
        linear_series_multi = linear_series.stack(linear_series)

        # univariate model + univariate series
        score = NaiveDrift().backtest(
            linear_series,
            start=pd.Timestamp("20000201"),
            forecast_horizon=3,
            metric=r2_score,
        )
        self.assertEqual(score, 1.0)

        # univariate model + univariate series + historical_forecasts precalculated
        forecasts = NaiveDrift().historical_forecasts(
            linear_series,
            start=pd.Timestamp("20000201"),
            forecast_horizon=3,
            last_points_only=False,
        )
        precalculated_forecasts_score = NaiveDrift().backtest(
            linear_series,
            historical_forecasts=forecasts,
            start=pd.Timestamp("20000201"),
            forecast_horizon=3,
            metric=r2_score,
        )
        self.assertEqual(score, precalculated_forecasts_score)

        # very large train length should not affect the backtest
        score = NaiveDrift().backtest(
            linear_series,
            train_length=10000,
            start=pd.Timestamp("20000201"),
            forecast_horizon=3,
            metric=r2_score,
        )
        self.assertEqual(score, 1.0)

        # using several metric function should not affect the backtest
        score = NaiveDrift().backtest(
            linear_series,
            train_length=10000,
            start=pd.Timestamp("20000201"),
            forecast_horizon=3,
            metric=[r2_score, mape],
        )
        np.testing.assert_almost_equal(score, np.array([1.0, 0.0]))

        # window of size 2 is too small for naive drift
        with self.assertRaises(ValueError):
            NaiveDrift().backtest(
                linear_series,
                train_length=2,
                start=pd.Timestamp("20000201"),
                forecast_horizon=3,
                metric=r2_score,
            )

        # test that it also works for time series that are not Datetime-indexed
        score = NaiveDrift().backtest(
            linear_series_int, start=0.7, forecast_horizon=3, metric=r2_score
        )
        self.assertEqual(score, 1.0)

        with self.assertRaises(ValueError):
            NaiveDrift().backtest(
                linear_series,
                start=pd.Timestamp("20000218"),
                forecast_horizon=3,
                overlap_end=False,
            )
        NaiveDrift().backtest(
            linear_series, start=pd.Timestamp("20000217"), forecast_horizon=3
        )
        NaiveDrift().backtest(
            linear_series,
            start=pd.Timestamp("20000218"),
            forecast_horizon=3,
            overlap_end=True,
        )

        # Using forecast_horizon default value
        NaiveDrift().backtest(linear_series, start=pd.Timestamp("20000216"))
        NaiveDrift().backtest(
            linear_series, start=pd.Timestamp("20000217"), overlap_end=True
        )

        # Using an int or float value for start
        NaiveDrift().backtest(linear_series, start=30)
        NaiveDrift().backtest(linear_series, start=0.7, overlap_end=True)

        # Set custom train window length
        NaiveDrift().backtest(linear_series, train_length=10, start=30)

        # Using invalid start and/or forecast_horizon values
        with self.assertRaises(ValueError):
            NaiveDrift().backtest(linear_series, start=0.7, forecast_horizon=-1)
        with self.assertRaises(ValueError):
            NaiveDrift().backtest(linear_series, start=-0.7, forecast_horizon=1)

        with self.assertRaises(ValueError):
            NaiveDrift().backtest(linear_series, start=100)
        with self.assertRaises(ValueError):
            NaiveDrift().backtest(linear_series, start=1.2)
        with self.assertRaises(TypeError):
            NaiveDrift().backtest(linear_series, start="wrong type")
        with self.assertRaises(ValueError):
            NaiveDrift().backtest(linear_series, train_length=0, start=0.5)
        with self.assertRaises(TypeError):
            NaiveDrift().backtest(linear_series, train_length=1.2, start=0.5)
        with self.assertRaises(TypeError):
            NaiveDrift().backtest(linear_series, train_length="wrong type", start=0.5)

        with self.assertRaises(ValueError):
            NaiveDrift().backtest(
                linear_series, start=49, forecast_horizon=2, overlap_end=False
            )

        # univariate model + multivariate series
        with self.assertRaises(ValueError):
            FFT().backtest(
                linear_series_multi, start=pd.Timestamp("20000201"), forecast_horizon=3
            )

        # multivariate model + univariate series
        if TORCH_AVAILABLE:
            tcn_model = TCNModel(
                input_chunk_length=12, output_chunk_length=1, batch_size=1, n_epochs=1
            )
            # cannot perform historical forecasts with `retrain=False` and untrained model
            with pytest.raises(ValueError):
                _ = tcn_model.historical_forecasts(
                    linear_series,
                    start=pd.Timestamp("20000125"),
                    forecast_horizon=3,
                    verbose=False,
                    last_points_only=True,
                    retrain=False,
                )

            pred = tcn_model.historical_forecasts(
                linear_series,
                start=pd.Timestamp("20000125"),
                forecast_horizon=3,
                verbose=False,
                last_points_only=True,
            )
            self.assertEqual(pred.width, 1)
            self.assertEqual(pred.end_time(), linear_series.end_time())

            # multivariate model + multivariate series
            # historical forecasts doesn't overwrite model object -> we can use different input dimensions
            tcn_model.backtest(
                linear_series_multi,
                start=pd.Timestamp("20000125"),
                forecast_horizon=3,
                verbose=False,
            )

            # univariate model
            tcn_model = TCNModel(
                input_chunk_length=12, output_chunk_length=1, batch_size=1, n_epochs=1
            )
            tcn_model.fit(linear_series, verbose=False)
            # univariate fitted model + multivariate series
            with pytest.raises(ValueError):
                tcn_model.backtest(
                    linear_series_multi,
                    start=pd.Timestamp("20000125"),
                    forecast_horizon=3,
                    verbose=False,
                    retrain=False,
                )

            tcn_model = TCNModel(
                input_chunk_length=12, output_chunk_length=3, batch_size=1, n_epochs=1
            )
            pred = tcn_model.historical_forecasts(
                linear_series_multi,
                start=pd.Timestamp("20000125"),
                forecast_horizon=3,
                verbose=False,
                last_points_only=True,
            )
            self.assertEqual(pred.width, 2)
            self.assertEqual(pred.end_time(), linear_series.end_time())

    def test_backtest_multiple_series(self):
        series = [AirPassengersDataset().load(), MonthlyMilkDataset().load()]
        model = NaiveSeasonal(K=1)

        error = model.backtest(
            series,
            train_length=30,
            forecast_horizon=2,
            stride=1,
            retrain=True,
            last_points_only=False,
            verbose=False,
        )

        expected = [11.63104, 6.09458]
        self.assertEqual(len(error), 2)
        self.assertAlmostEqual(error[0], expected[0], places=4)
        self.assertAlmostEqual(error[1], expected[1], places=4)

    @unittest.skipUnless(TORCH_AVAILABLE, "requires torch")
    def test_backtest_regression(self):
        np.random.seed(4)

        gaussian_series = gt(mean=2, length=50)
        sine_series = st(length=50)
        features = gaussian_series.stack(sine_series)
        features_multivariate = (
            (gaussian_series + sine_series).stack(gaussian_series).stack(sine_series)
        )
        target = sine_series

        features = features.with_columns_renamed(
            features.components, ["Value0", "Value1"]
        )

        features_multivariate = features_multivariate.with_columns_renamed(
            features_multivariate.components, ["Value0", "Value1", "Value2"]
        )

        # univariate feature test
        score = LinearRegressionModel(
            lags=None, lags_future_covariates=[0, -1]
        ).backtest(
            series=target,
            future_covariates=features,
            start=pd.Timestamp("20000201"),
            forecast_horizon=3,
            metric=r2_score,
            last_points_only=True,
        )
        self.assertGreater(score, 0.9)

        # univariate feature test + train length
        score = LinearRegressionModel(
            lags=None, lags_future_covariates=[0, -1]
        ).backtest(
            series=target,
            future_covariates=features,
            start=pd.Timestamp("20000201"),
            train_length=20,
            forecast_horizon=3,
            metric=r2_score,
            last_points_only=True,
        )
        self.assertGreater(score, 0.9)

        # Using an int or float value for start
        score = RandomForest(
            lags=12, lags_future_covariates=[0], random_state=0
        ).backtest(
            series=target,
            future_covariates=features,
            start=30,
            forecast_horizon=3,
            metric=r2_score,
        )
        self.assertGreater(score, 0.9)

        score = RandomForest(
            lags=12, lags_future_covariates=[0], random_state=0
        ).backtest(
            series=target,
            future_covariates=features,
            start=0.5,
            forecast_horizon=3,
            metric=r2_score,
        )
        self.assertGreater(score, 0.9)

        # Using a too small start value
        with self.assertRaises(ValueError):
            RandomForest(lags=12).backtest(series=target, start=0, forecast_horizon=3)

        with self.assertRaises(ValueError):
            RandomForest(lags=12).backtest(
                series=target, start=0.01, forecast_horizon=3
            )

        # Using RandomForest's start default value
        score = RandomForest(lags=12, random_state=0).backtest(
            series=target, forecast_horizon=3, start=0.5, metric=r2_score
        )
        self.assertGreater(score, 0.95)

        # multivariate feature test
        score = RandomForest(
            lags=12, lags_future_covariates=[0, -1], random_state=0
        ).backtest(
            series=target,
            future_covariates=features_multivariate,
            start=pd.Timestamp("20000201"),
            forecast_horizon=3,
            metric=r2_score,
        )
        self.assertGreater(score, 0.94)

        # multivariate feature test with train window 35
        score_35 = RandomForest(
            lags=12, lags_future_covariates=[0, -1], random_state=0
        ).backtest(
            series=target,
            train_length=35,
            future_covariates=features_multivariate,
            start=pd.Timestamp("20000201"),
            forecast_horizon=3,
            metric=r2_score,
        )
        logger.info(
            "Score for multivariate feature test with train window 35 is: ", score_35
        )
        self.assertGreater(score_35, 0.92)

        # multivariate feature test with train window 45
        score_45 = RandomForest(
            lags=12, lags_future_covariates=[0, -1], random_state=0
        ).backtest(
            series=target,
            train_length=45,
            future_covariates=features_multivariate,
            start=pd.Timestamp("20000201"),
            forecast_horizon=3,
            metric=r2_score,
        )
        logger.info(
            "Score for multivariate feature test with train window 45 is: ", score_45
        )
        self.assertGreater(score_45, 0.94)
        self.assertGreater(score_45, score_35)

        # multivariate with stride
        score = RandomForest(
            lags=12, lags_future_covariates=[0], random_state=0
        ).backtest(
            series=target,
            future_covariates=features_multivariate,
            start=pd.Timestamp("20000201"),
            forecast_horizon=3,
            metric=r2_score,
            last_points_only=True,
            stride=3,
        )
        self.assertGreater(score, 0.9)

    def test_gridsearch(self):
        np.random.seed(1)

        dummy_series = get_dummy_series(ts_length=50)
        dummy_series_int_index = TimeSeries.from_values(dummy_series.values())

        theta_params = {"theta": list(range(3, 10))}
        self.assertTrue(compare_best_against_random(Theta, theta_params, dummy_series))
        self.assertTrue(
            compare_best_against_random(Theta, theta_params, dummy_series_int_index)
        )
        self.assertTrue(
            compare_best_against_random(Theta, theta_params, dummy_series, stride=2)
        )

        fft_params = {"nr_freqs_to_keep": [10, 50, 100], "trend": [None, "poly", "exp"]}
        self.assertTrue(compare_best_against_random(FFT, fft_params, dummy_series))

        es_params = {"seasonal_periods": list(range(5, 10))}
        self.assertTrue(
            compare_best_against_random(ExponentialSmoothing, es_params, dummy_series)
        )

    def test_gridsearch_metric_score(self):
        np.random.seed(1)

        model_class = Theta
        params = {"theta": list(range(3, 6))}
        dummy_series = get_dummy_series(ts_length=50)

        best_model, _, score = model_class.gridsearch(
            params,
            series=dummy_series,
            forecast_horizon=10,
            stride=1,
            start=dummy_series.time_index[-21],
        )
        recalculated_score = best_model.backtest(
            series=dummy_series,
            start=dummy_series.time_index[-21],
            forecast_horizon=10,
            stride=1,
        )

        self.assertEqual(score, recalculated_score, "The metric scores should match")

    @unittest.skipUnless(TORCH_AVAILABLE, "requires torch")
    def test_gridsearch_random_search(self):
        np.random.seed(1)

        dummy_series = get_dummy_series(ts_length=50)

        param_range = list(range(10, 20))
        params = {"lags": param_range}

        model = RandomForest(lags=1)
        result = model.gridsearch(
            params, dummy_series, forecast_horizon=1, n_random_samples=5
        )

        self.assertEqual(type(result[0]), RandomForest)
        self.assertEqual(type(result[1]["lags"]), int)
        self.assertEqual(type(result[2]), float)
        self.assertTrue(min(param_range) <= result[1]["lags"] <= max(param_range))

    @unittest.skipUnless(TORCH_AVAILABLE, "requires torch")
    def test_gridsearch_n_random_samples_bad_arguments(self):
        dummy_series = get_dummy_series(ts_length=50)

        params = {"lags": list(range(1, 11)), "past_covariates": list(range(1, 11))}

        with self.assertRaises(ValueError):
            RandomForest.gridsearch(
                params, dummy_series, forecast_horizon=1, n_random_samples=-5
            )
        with self.assertRaises(ValueError):
            RandomForest.gridsearch(
                params, dummy_series, forecast_horizon=1, n_random_samples=105
            )
        with self.assertRaises(ValueError):
            RandomForest.gridsearch(
                params, dummy_series, forecast_horizon=1, n_random_samples=-24.56
            )
        with self.assertRaises(ValueError):
            RandomForest.gridsearch(
                params, dummy_series, forecast_horizon=1, n_random_samples=1.5
            )

    @unittest.skipUnless(TORCH_AVAILABLE, "requires torch")
    def test_gridsearch_n_random_samples(self):
        np.random.seed(1)

        params = {"lags": list(range(1, 11)), "past_covariates": list(range(1, 11))}

        params_cross_product = list(product(*params.values()))

        # Test absolute sample
        absolute_sampled_result = RandomForest._sample_params(params_cross_product, 10)
        self.assertEqual(len(absolute_sampled_result), 10)

        # Test percentage sample
        percentage_sampled_result = RandomForest._sample_params(
            params_cross_product, 0.37
        )
        self.assertEqual(len(percentage_sampled_result), 37)

    @unittest.skipUnless(TORCH_AVAILABLE, "requires torch")
    def test_gridsearch_n_jobs(self):
        """
        Testing that running gridsearch with multiple workers returns the same
        best_parameters as the single worker run.
        """

        np.random.seed(1)

        dummy_series = get_dummy_series(
            ts_length=100, lt_end_value=1, st_value_offset=0
        ).astype(np.float32)
        ts_train, ts_val = dummy_series.split_before(split_point=0.8)

        test_cases = [
            {
                "model": ARIMA,  # ExtendedForecastingModel
                "parameters": {"p": [18, 4], "q": [2, 3]},
            },
            {
                "model": BlockRNNModel,  # TorchForecastingModel
                "parameters": {
                    "input_chunk_length": [5, 10],
                    "output_chunk_length": [1, 3],
                    "n_epochs": [1, 5],
                    "random_state": [
                        42
                    ],  # necessary to avoid randomness among runs with same parameters
                },
            },
        ]

        for test in test_cases:

            model = test["model"]
            parameters = test["parameters"]

            np.random.seed(1)
            _, best_params1, _ = model.gridsearch(
                parameters=parameters, series=ts_train, val_series=ts_val, n_jobs=1
            )

            np.random.seed(1)
            _, best_params2, _ = model.gridsearch(
                parameters=parameters, series=ts_train, val_series=ts_val, n_jobs=-1
            )

            self.assertEqual(best_params1, best_params2)

    @unittest.skipUnless(TORCH_AVAILABLE, "requires torch")
    def test_gridsearch_multi(self):
        dummy_series = st(length=40, value_y_offset=10).stack(
            lt(length=40, end_value=20)
        )
        tcn_params = {
            "input_chunk_length": [12],
            "output_chunk_length": [3],
            "n_epochs": [1],
            "batch_size": [1],
            "kernel_size": [2, 3, 4],
        }
        TCNModel.gridsearch(tcn_params, dummy_series, forecast_horizon=3, metric=mape)



"""
Facebook Prophet
----------------
"""

import logging
import re
from typing import Callable, List, Optional, Sequence, Union

import numpy as np
import pandas as pd
import prophet

from darts.logging import execute_and_suppress_output, get_logger, raise_if
from darts.models.forecasting.forecasting_model import (
    FutureCovariatesLocalForecastingModel,
)
from darts.timeseries import TimeSeries

logger = get_logger(__name__)
logger.level = logging.WARNING  # set to warning to suppress prophet logs


class Prophet(FutureCovariatesLocalForecastingModel):
    def __init__(
        self,
        add_seasonalities: Optional[Union[dict, List[dict]]] = None,
        country_holidays: Optional[str] = None,
        suppress_stdout_stderror: bool = True,
        add_encoders: Optional[dict] = None,
        cap: Union[
            float, Callable[[Union[pd.DatetimeIndex, pd.RangeIndex]], Sequence[float]]
        ] = None,
        floor: Union[
            float, Callable[[Union[pd.DatetimeIndex, pd.RangeIndex]], Sequence[float]]
        ] = None,
        **prophet_kwargs,
    ):
        """Facebook Prophet

        This class provides a basic wrapper around `Facebook Prophet <https://github.com/facebook/prophet>`_.
        It supports adding country holidays as well as custom seasonalities and adds support for stochastic
        forecasting and future covariates.

        Parameters
        ----------
        add_seasonalities
            Optionally, a dict or list of dicts with custom seasonality/ies to add to the model.
            Each dict takes the following mandatory and optional data:

            .. highlight:: python
            .. code-block:: python

                dict({
                'name': str  # (name of the seasonality component),
                'seasonal_periods': int  # (nr of steps composing a season),
                'fourier_order': int  # (number of Fourier components to use),
                'prior_scale': Optional[float]  # (a prior scale for this component),
                'mode': Optional[str]  # ('additive' or 'multiplicative')
                })
            ..

            An example for `seasonal_periods`: If you have hourly data (frequency='H') and your seasonal cycle repeats
            after 48 hours then set `seasonal_periods=48`.

            Apart from `seasonal_periods`, this is very similar to how you would call Facebook Prophet's
            `add_seasonality()` method.
            Alternatively, you can add seasonalities after model creation and before fitting with
            :meth:`add_seasonality() <Prophet.add_seasonality()>`.
        country_holidays
            An optional country code, for which holidays can be taken into account by Prophet.

            See: https://github.com/dr-prodigy/python-holidays

            In addition to those countries, Prophet includes holidays for these
            countries: Brazil (BR), Indonesia (ID), India (IN), Malaysia (MY), Vietnam (VN),
            Thailand (TH), Philippines (PH), Turkey (TU), Pakistan (PK), Bangladesh (BD),
            Egypt (EG), China (CN), and Russia (RU).
        suppress_stdout_stderror
            Optionally suppress the log output produced by Prophet during training.
        add_encoders
            A large number of future covariates can be automatically generated with `add_encoders`.
            This can be done by adding multiple pre-defined index encoders and/or custom user-made functions that
            will be used as index encoders. Additionally, a transformer such as Darts' :class:`Scaler` can be added to
            transform the generated covariates. This happens all under one hood and only needs to be specified at
            model creation.
            Read :meth:`SequentialEncoder <darts.dataprocessing.encoders.SequentialEncoder>` to find out more about
            ``add_encoders``. Default: ``None``. An example showing some of ``add_encoders`` features:

            .. highlight:: python
            .. code-block:: python

                add_encoders={
                    'cyclic': {'future': ['month']},
                    'datetime_attribute': {'future': ['hour', 'dayofweek']},
                    'position': {'future': ['relative']},
                    'custom': {'future': [lambda idx: (idx.year - 1950) / 50]},
                    'transformer': Scaler()
                }
            ..
        cap
            Parameter specifiying the maximum carrying capacity when predicting with logistic growth.
            Mandatory when `growth = 'logistic'`, otherwise ignored.
            See <https://facebook.github.io/prophet/docs/saturating_forecasts.html> for more information
            on logistic forecasts.
            Can be either

            - a number, for constant carrying capacities
            - a function taking a DatetimeIndex or RangeIndex and returning a corresponding a Sequence of numbers,
            where each number indicates the carrying capacity at this index.
        floor
            Parameter specifiying the minimum carrying capacity when predicting logistic growth.
            Optional when `growth = 'logistic'` (defaults to 0), otherwise ignored.
            See <https://facebook.github.io/prophet/docs/saturating_forecasts.html> for more information
            on logistic forecasts.
            Can be either

            - a number, for constant carrying capacities
            - a function taking a DatetimeIndex or RangeIndex and returning a corresponding a Sequence of numbers,
            where each number indicates the carrying capacity at this index.
        prophet_kwargs
            Some optional keyword arguments for Prophet.
            For information about the parameters see:
            `The Prophet source code <https://github.com/facebook/prophet/blob/master/python/prophet/forecaster.py>`_.
        """

        super().__init__(add_encoders=add_encoders)

        self._auto_seasonalities = self._extract_auto_seasonality(prophet_kwargs)

        self._add_seasonalities = dict()
        add_seasonality_calls = (
            add_seasonalities
            if isinstance(add_seasonalities, list)
            else [add_seasonalities]
        )
        for call in add_seasonality_calls:
            self._store_add_seasonality_call(seasonality_call=call)

        self.country_holidays = country_holidays
        self.prophet_kwargs = prophet_kwargs
        self.model = None
        self.suppress_stdout_stderr = suppress_stdout_stderror

        self._execute_and_suppress_output = execute_and_suppress_output
        self._model_builder = prophet.Prophet

        self._cap = cap
        self._floor = floor
        self.is_logistic = (
            "growth" in prophet_kwargs and prophet_kwargs["growth"] == "logistic"
        )
        if not self.is_logistic and (cap is not None or floor is not None):
            logger.warning(
                "Parameters `cap` and/or `floor` were set although `growth` is not "
                "logistic. The set capacities will be ignored."
            )
        if self.is_logistic:
            raise_if(
                cap is None,
                "Parameter `cap` has to be set when `growth` is logistic",
                logger,
            )
            if floor is None:
                # Use 0 as default value
                self._floor = 0

    def _fit(self, series: TimeSeries, future_covariates: Optional[TimeSeries] = None):

        super()._fit(series, future_covariates)
        self._assert_univariate(series)
        series = self.training_series

        fit_df = pd.DataFrame(
            data={"ds": series.time_index, "y": series.univariate_values()}
        )
        if self.is_logistic:
            fit_df = self._add_capacities_to_df(fit_df)

        self.model = self._model_builder(**self.prophet_kwargs)

        # add user defined seasonalities (from model creation and/or pre-fit self.add_seasonalities())
        interval_length = self._freq_to_days(series.freq_str)
        for seasonality_name, attributes in self._add_seasonalities.items():
            self.model.add_seasonality(
                name=seasonality_name,
                period=attributes["seasonal_periods"] * interval_length,
                fourier_order=attributes["fourier_order"],
            )

        # add covariates
        if future_covariates is not None:
            fit_df = fit_df.merge(
                future_covariates.pd_dataframe(),
                left_on="ds",
                right_index=True,
                how="left",
            )
            for covariate in future_covariates.columns:
                self.model.add_regressor(covariate)

        # add built-in country holidays
        if self.country_holidays is not None:
            self.model.add_country_holidays(self.country_holidays)

        if self.suppress_stdout_stderr:
            self._execute_and_suppress_output(
                self.model.fit, logger, logging.WARNING, fit_df
            )
        else:
            self.model.fit(fit_df)

        return self

    def _predict(
        self,
        n: int,
        future_covariates: Optional[TimeSeries] = None,
        num_samples: int = 1,
        verbose: bool = False,
    ) -> TimeSeries:

        super()._predict(n, future_covariates, num_samples)

        predict_df = self._generate_predict_df(n=n, future_covariates=future_covariates)

        if num_samples == 1:
            forecast = self.model.predict(predict_df, vectorized=True)["yhat"].values
        else:
            forecast = np.expand_dims(
                self._stochastic_samples(predict_df, n_samples=num_samples), axis=1
            )

        return self._build_forecast_series(forecast)

    def _add_capacities_to_df(self, df: pd.DataFrame) -> pd.DataFrame:
        dates = df["ds"]
        try:
            df["cap"] = self._cap(dates) if callable(self._cap) else self._cap
            df["floor"] = self._floor(dates) if callable(self._floor) else self._floor
        except ValueError as e:
            raise_if(
                "does not match length of index" in str(e),
                "Callables supplied to `Prophet.set_capacity` as `cap` or `floor` "
                "arguments have to return Sequences of identical length as their "
                " input argument Sequence!",
                logger,
            )
            raise
        return df

    def _generate_predict_df(
        self, n: int, future_covariates: Optional[TimeSeries] = None
    ) -> pd.DataFrame:
        """Returns a pandas DataFrame in the format required for Prophet.predict() with `n` dates after the end of
        the fitted TimeSeries"""

        predict_df = pd.DataFrame(data={"ds": self._generate_new_dates(n)})
        if self.is_logistic:
            predict_df = self._add_capacities_to_df(predict_df)
        if future_covariates is not None:
            predict_df = predict_df.merge(
                future_covariates.pd_dataframe(),
                left_on="ds",
                right_index=True,
                how="left",
            )
        return predict_df

    def _is_probabilistic(self) -> bool:
        return True

    def _stochastic_samples(self, predict_df, n_samples) -> np.ndarray:
        """Returns stochastic forecast of `n_samples` samples.
        This method is a replicate of Prophet.predict() which suspends simplification of stochastic samples to
        deterministic target values."""

        # save default number of uncertainty_samples and set user-defined n_samples
        n_samples_default = self.model.uncertainty_samples
        self.model.uncertainty_samples = n_samples

        if self.model.history is None:
            raise ValueError("Model has not been fit.")

        if predict_df is None:
            predict_df = self.model.history.copy()
        else:
            if predict_df.shape[0] == 0:
                raise ValueError("Dataframe has no rows.")
            predict_df = self.model.setup_dataframe(predict_df.copy())

        predict_df["trend"] = self.model.predict_trend(predict_df)

        forecast = self.model.sample_posterior_predictive(predict_df, vectorized=True)

        # reset default number of uncertainty_samples
        self.model.uncertainty_samples = n_samples_default
        return forecast["yhat"]

    def predict_raw(
        self, n: int, future_covariates: Optional[TimeSeries] = None
    ) -> pd.DataFrame:
        """Returns the output of the base Facebook Prophet model in form of a pandas DataFrame. Note however,
        that the output of this method is not supported for further processing with the Darts API.

        Methods of the base Prophet model can be accessed with self.model.method() (i.e. self.model.plot_components())
        """
        super().predict(n, future_covariates, num_samples=1)

        predict_df = self._generate_predict_df(n=n, future_covariates=future_covariates)

        return self.model.predict(predict_df, vectorized=True)

    def add_seasonality(
        self,
        name: str,
        seasonal_periods: int,
        fourier_order: int,
        prior_scale: Optional[float] = None,
        mode: Optional[str] = None,
    ) -> None:
        """Adds a custom seasonality to the model that repeats after every n `seasonal_periods` timesteps.
        An example for `seasonal_periods`: If you have hourly data (frequency='H') and your seasonal cycle repeats
        after 48 hours -> `seasonal_periods=48`.

        Apart from `seasonal_periods`, this is very similar to how you would call Facebook Prophet's
        `add_seasonality()` method. For information about the parameters see:
        `The Prophet source code <https://github.com/facebook/prophet/blob/master/python/prophet/forecaster.py>`_.

        Parameters
        ----------
        name
            name of the seasonality component
        seasonal_periods
            number of timesteps after which the seasonal cycle repeats
        fourier_order
            number of Fourier components to use
        prior_scale
            optionally, a prior scale for this component
        mode
            optionally, 'additive' or 'multiplicative'
        """
        function_call = {
            "name": name,
            "seasonal_periods": seasonal_periods,
            "fourier_order": fourier_order,
            "prior_scale": prior_scale,
            "mode": mode,
        }
        self._store_add_seasonality_call(seasonality_call=function_call)

    def _store_add_seasonality_call(
        self, seasonality_call: Optional[dict] = None
    ) -> None:
        """Checks the validity of an add_seasonality() call and stores valid calls.
        As the actual model is only created at fitting time, and seasonalities are added pre-fit,
        the add_seasonality calls must be stored and checked on Darts' side.

        Raises
        ----------
        ValueError
            if `seasonality_call` has missing or empty mandatory keys/arguments

            if `seasonality_call` with `name` already exists.

            if `seasonality_call` has invalid keys/arguments

            if `seasonality_call` has invalid dtypes
        """

        if seasonality_call is None:
            return

        seasonality_properties = {
            "name": {"default": None, "dtype": str},
            "seasonal_periods": {"default": None, "dtype": int},
            "fourier_order": {"default": None, "dtype": int},
            "prior_scale": {"default": None, "dtype": float},
            "mode": {"default": None, "dtype": str},
        }
        seasonality_default = {
            kw: seasonality_properties[kw]["default"] for kw in seasonality_properties
        }

        mandatory_keywords = ["name", "seasonal_periods", "fourier_order"]

        add_seasonality_call = dict(seasonality_default, **seasonality_call)

        missing_kws = [
            kw for kw in mandatory_keywords if add_seasonality_call[kw] is None
        ]
        raise_if(
            len(missing_kws) > 0,
            f'Seasonality `{add_seasonality_call["name"]}` has missing mandatory keywords or empty arguments: '
            f"{missing_kws}.",
            logger,
        )

        seasonality_name = add_seasonality_call["name"]
        raise_if(
            seasonality_name in self._auto_seasonalities
            or seasonality_name in self._add_seasonalities,
            f"Adding seasonality with `name={seasonality_name}` failed. A seasonality with this name already "
            f"exists.",
        )

        invalid_kws = [
            kw for kw in add_seasonality_call if kw not in seasonality_default
        ]
        raise_if(
            len(invalid_kws) > 0,
            f'Seasonality `{add_seasonality_call["name"]}` has invalid keywords: {invalid_kws}. Only the '
            f"following arguments are supported: {list(seasonality_default)}",
            logger,
        )

        invalid_types = [
            kw
            for kw, value in add_seasonality_call.items()
            if not isinstance(value, seasonality_properties[kw]["dtype"])
            and value is not None
        ]
        raise_if(
            len(invalid_types) > 0,
            f'Seasonality `{add_seasonality_call["name"]}` has invalid value dtypes: {invalid_types} must be '
            f'of type {[seasonality_properties[kw]["dtype"] for kw in invalid_types]}.',
            logger,
        )
        self._add_seasonalities[seasonality_name] = add_seasonality_call

    @staticmethod
    def _extract_auto_seasonality(prophet_kwargs: dict) -> list:
        """Returns the automatically added seasonalities by Prophet's base model based on kwargs of model creation"""
        auto_seasonalities = []
        for auto_seasonality in ["daily", "weekly", "yearly"]:
            s_name = auto_seasonality + "_seasonality"
            if not (s_name in prophet_kwargs and not prophet_kwargs[s_name]):
                auto_seasonalities.append(auto_seasonality)
        return auto_seasonalities

    @staticmethod
    def _freq_to_days(freq: str) -> float:
        """Converts a frequency to number of days required by Facebook Prophet

        Parameters
        ----------
        freq
            frequency string of the underlying TimeSeries's time index (pd.DateTimeIndex.freq_str)
        """

        # this regex extracts all digits from `freq`: exp: '30S' -> 30
        freq_times = re.findall(r"\d+", freq)
        freq_times = 1 if not freq_times else int(freq_times[0])

        # this regex extracts all characters and '-' from `freq` and then extracts left string from '-'
        # exp: 'W-SUN' -> 'W', '30S' -> 'S'
        freq = "".join(re.split("[^a-zA-Z-]*", freq)).split("-")[0]

        seconds_per_day = 86400
        if freq in ["A", "BA", "Y", "BY", "RE"] or freq.startswith(
            ("A", "BA", "Y", "BY", "RE")
        ):  # year
            days = 365.25
        elif freq in ["Q", "BQ", "REQ"] or freq.startswith(
            ("Q", "BQ", "REQ")
        ):  # quarter
            days = 3 * 30.4375
        elif freq in ["M", "BM", "CBM", "SM"] or freq.startswith(
            ("M", "BM", "BS", "CBM", "SM")
        ):  # month
            days = 30.4375
        elif freq in ["W"]:  # week
            days = 7.0
        elif freq in ["B", "C"]:  # business day
            days = 1 * 7 / 5
        elif freq in ["D"]:  # day
            days = 1.0
        elif freq in ["H", "BH", "CBH"]:  # hour
            days = 1 / 24
        elif freq in ["T", "min"]:  # minute
            days = 1 / (24 * 60)
        elif freq in ["S"]:  # second
            days = 1 / seconds_per_day
        elif freq in ["L", "ms"]:  # millisecond
            days = 1 / (seconds_per_day * 10**3)
        elif freq in ["U", "us"]:  # microsecond
            days = 1 / (seconds_per_day * 10**6)
        elif freq in ["N"]:  # nanosecond
            days = 1 / (seconds_per_day * 10**9)
        else:
            raise ValueError(
                "freq {} not understood. Please report if you think this is in error.".format(
                    freq
                )
            )
        return freq_times * days

    def _supports_range_index(self) -> bool:
        """Prophet does not support integer range index."""
        raise_if(
            True,
            "Prophet does not support integer range index. The index of the TimeSeries must be of type "
            "pandas.DatetimeIndex",
            logger,
        )
        return False



"""
PyTorch Loss Functions
----------------------
"""
# Inspiration: https://github.com/ElementAI/N-BEATS/blob/master/common/torch/losses.py

import numpy as np
import torch
import torch.nn as nn


def _divide_no_nan(a, b):
    """
    a/b where the resulted NaN or Inf are replaced by 0.
    """
    result = a / b
    result[result != result] = 0.0
    result[result == np.inf] = 0.0
    result[result == np.NINF] = 0.0
    return result


class SmapeLoss(nn.Module):
    def __init__(self, block_denom_grad: bool = True):
        """
        sMAPE loss as defined in https://robjhyndman.com/hyndsight/smape/ (Chen and Yang 2004)

        Given a time series of actual values :math:`y_t` and a time series of predicted values :math:`\\hat{y}_t`
        both of length :math:`T`, it is computed as

        .. math::
            \\frac{1}{T}
            \\sum_{t=1}^{T}{\\frac{\\left| y_t - \\hat{y}_t \\right|}
                                  {\\left| y_t \\right| + \\left| \\hat{y}_t \\right|} }.

        The results of divisions yielding NaN or Inf are replaced by 0. Note that we drop the coefficient of
        200 usually used for computing sMAPE values, as it impacts only the magnitude of the gradients
        and not their direction.

        Parameters
        ----------
        block_denom_grad
            Whether to stop the gradient in the denominator
        """
        super().__init__()
        self.block_denom_grad = block_denom_grad

    def forward(self, inpt, tgt):
        num = torch.abs(tgt - inpt)
        denom = torch.abs(tgt) + torch.abs(inpt)
        if self.block_denom_grad:
            denom = denom.detach()
        return torch.mean(_divide_no_nan(num, denom))


class MapeLoss(nn.Module):
    def __init__(self):
        """
        MAPE loss as defined in: https://en.wikipedia.org/wiki/Mean_absolute_percentage_error.

        Given a time series of actual values :math:`y_t` and a time series of predicted values :math:`\\hat{y}_t`
        both of length :math:`T`, it is computed as

        .. math::
            \\frac{1}{T}
            \\sum_{t=1}^{T}{\\frac{\\left| y_t - \\hat{y}_t \\right|}{y_t}}.

        The results of divisions yielding NaN or Inf are replaced by 0. Note that we drop the coefficient of
        100 usually used for computing MAPE values, as it impacts only the magnitude of the gradients
        and not their direction.
        """
        super().__init__()

    def forward(self, inpt, tgt):
        return torch.mean(torch.abs(_divide_no_nan(tgt - inpt, tgt)))


class MAELoss(nn.Module):
    def __init__(self):
        """
        MAE loss as defined in: https://en.wikipedia.org/wiki/Mean_absolute_error.

        Given a time series of actual values :math:`y_t` and a time series of predicted values :math:`\\hat{y}_t`
        both of length :math:`T`, it is computed as

        .. math::
            \\frac{1}{T}
            \\sum_{t=1}^{T}{\\left| y_t - \\hat{y}_t \\right|}.

        Note that this is the same as torch.nn.L1Loss.
        """
        super().__init__()

    def forward(self, inpt, tgt):
        return torch.mean(torch.abs(tgt - inpt))



"""
ARIMA
-----

Models for ARIMA (Autoregressive integrated moving average) [1]_.
The implementations is wrapped around `statsmodels <https://github.com/statsmodels/statsmodels>`_.

References
----------
.. [1] https://wikipedia.org/wiki/Autoregressive_integrated_moving_average
"""

from typing import Optional, Tuple

import numpy as np
from statsmodels.tsa.arima.model import ARIMA as staARIMA

from darts.logging import get_logger
from darts.models.forecasting.forecasting_model import (
    TransferableFutureCovariatesLocalForecastingModel,
)
from darts.timeseries import TimeSeries

logger = get_logger(__name__)


class ARIMA(TransferableFutureCovariatesLocalForecastingModel):
    def __init__(
        self,
        p: int = 12,
        d: int = 1,
        q: int = 0,
        seasonal_order: Tuple[int, int, int, int] = (0, 0, 0, 0),
        trend: Optional[str] = None,
        random_state: int = 0,
        add_encoders: Optional[dict] = None,
    ):
        """ARIMA
        ARIMA-type models extensible with exogenous variables (future covariates)
        and seasonal components.

        Parameters
        ----------
        p : int
            Order (number of time lags) of the autoregressive model (AR).
        d : int
            The order of differentiation; i.e., the number of times the data
            have had past values subtracted (I).
        q : int
            The size of the moving average window (MA).
        seasonal_order: Tuple[int, int, int, int]
            The (P,D,Q,s) order of the seasonal component for the AR parameters,
            differences, MA parameters and periodicity.
        trend: str
            Parameter controlling the deterministic trend. 'n' indicates no trend,
            'c' a constant term, 't' linear trend in time, and 'ct' includes both.
            Default is 'c' for models without integration, and no trend for models with integration.
        add_encoders
            A large number of future covariates can be automatically generated with `add_encoders`.
            This can be done by adding multiple pre-defined index encoders and/or custom user-made functions that
            will be used as index encoders. Additionally, a transformer such as Darts' :class:`Scaler` can be added to
            transform the generated covariates. This happens all under one hood and only needs to be specified at
            model creation.
            Read :meth:`SequentialEncoder <darts.dataprocessing.encoders.SequentialEncoder>` to find out more about
            ``add_encoders``. Default: ``None``. An example showing some of ``add_encoders`` features:

            .. highlight:: python
            .. code-block:: python

                add_encoders={
                    'cyclic': {'future': ['month']},
                    'datetime_attribute': {'future': ['hour', 'dayofweek']},
                    'position': {'future': ['relative']},
                    'custom': {'future': [lambda idx: (idx.year - 1950) / 50]},
                    'transformer': Scaler()
                }
            ..
        """
        super().__init__(add_encoders=add_encoders)
        self.order = p, d, q
        self.seasonal_order = seasonal_order
        self.trend = trend
        self.model = None
        np.random.seed(random_state)

    def _fit(self, series: TimeSeries, future_covariates: Optional[TimeSeries] = None):
        super()._fit(series, future_covariates)

        self._assert_univariate(series)

        # storing to restore the statsmodels model results object
        self.training_historic_future_covariates = future_covariates

        m = staARIMA(
            series.values(copy=False),
            exog=future_covariates.values(copy=False) if future_covariates else None,
            order=self.order,
            seasonal_order=self.seasonal_order,
            trend=self.trend,
        )
        self.model = m.fit()

        return self

    def _predict(
        self,
        n: int,
        series: Optional[TimeSeries] = None,
        historic_future_covariates: Optional[TimeSeries] = None,
        future_covariates: Optional[TimeSeries] = None,
        num_samples: int = 1,
        verbose: bool = False,
    ) -> TimeSeries:

        if num_samples > 1 and self.trend:
            logger.warning(
                "Trends are not well supported yet for getting probabilistic forecasts with ARIMA."
                "If you run into issues, try calling fit() with num_samples=1 or removing the trend from"
                "your model."
            )

        super()._predict(
            n, series, historic_future_covariates, future_covariates, num_samples
        )

        # updating statsmodels results object state with the new ts and covariates
        if series is not None:
            self.model = self.model.apply(
                series.values(copy=False),
                exog=historic_future_covariates.values(copy=False)
                if historic_future_covariates
                else None,
            )

        if num_samples == 1:
            forecast = self.model.forecast(
                steps=n,
                exog=future_covariates.values(copy=False)
                if future_covariates
                else None,
            )
        else:
            forecast = self.model.simulate(
                nsimulations=n,
                repetitions=num_samples,
                initial_state=self.model.states.predicted[-1, :],
                exog=future_covariates.values(copy=False)
                if future_covariates
                else None,
            )

        # restoring statsmodels results object state
        if series is not None:
            self.model = self.model.apply(
                self._orig_training_series.values(copy=False),
                exog=self.training_historic_future_covariates.values(copy=False)
                if self.training_historic_future_covariates
                else None,
            )

        return self._build_forecast_series(forecast)

    def _is_probabilistic(self) -> bool:
        return True

    @property
    def min_train_series_length(self) -> int:
        return 30



"""
VARIMA
-----

Models for VARIMA (Vector Autoregressive moving average) [1]_.
The implementations is wrapped around `statsmodels <https://github.com/statsmodels/statsmodels>`_.

References
----------
.. [1] https://en.wikipedia.org/wiki/Vector_autoregression
"""
from typing import Optional

import numpy as np
import pandas as pd
from statsmodels.tsa.api import VARMAX as staVARMA

from darts.logging import get_logger, raise_if
from darts.models.forecasting.forecasting_model import (
    TransferableFutureCovariatesLocalForecastingModel,
)
from darts.timeseries import TimeSeries

logger = get_logger(__name__)


class VARIMA(TransferableFutureCovariatesLocalForecastingModel):
    def __init__(
        self,
        p: int = 1,
        d: int = 0,
        q: int = 0,
        trend: Optional[str] = None,
        add_encoders: Optional[dict] = None,
    ):
        """VARIMA

        Parameters
        ----------
        p : int
            Order (number of time lags) of the autoregressive model (AR)
        d : int
            The order of differentiation; i.e., the number of times the data
            have had past values subtracted. (I) Note that Darts only supports d <= 1 because for
            d > 1 the optimizer often does not result in stable predictions. If results are not stable
            for d = 1 try to set d = 0 and enable the trend parameter
            to account for possible non-stationarity.
        q : int
            The size of the moving average window (MA).
        trend: str
            Parameter controlling the deterministic trend. 'n' indicates no trend,
            'c' a constant term, 't' linear trend in time, and 'ct' includes both.
            Default is 'c' for models without integration, and no trend for models with integration.
        add_encoders
            A large number of future covariates can be automatically generated with `add_encoders`.
            This can be done by adding multiple pre-defined index encoders and/or custom user-made functions that
            will be used as index encoders. Additionally, a transformer such as Darts' :class:`Scaler` can be added to
            transform the generated covariates. This happens all under one hood and only needs to be specified at
            model creation.
            Read :meth:`SequentialEncoder <darts.dataprocessing.encoders.SequentialEncoder>` to find out more about
            ``add_encoders``. Default: ``None``. An example showing some of ``add_encoders`` features:

            .. highlight:: python
            .. code-block:: python

                add_encoders={
                    'cyclic': {'future': ['month']},
                    'datetime_attribute': {'future': ['hour', 'dayofweek']},
                    'position': {'future': ['relative']},
                    'custom': {'future': [lambda idx: (idx.year - 1950) / 50]},
                    'transformer': Scaler()
                }
            ..
        """
        super().__init__(add_encoders=add_encoders)
        self.p = p
        self.d = d
        self.q = q
        self.trend = trend
        self.model = None

        assert d <= 1, "d > 1 not supported."

    def _differentiate_series(self, series: TimeSeries) -> TimeSeries:
        """Differentiate the series self.d times"""
        for _ in range(self.d):
            series = TimeSeries.from_dataframe(
                df=series.pd_dataframe(copy=False).diff().dropna(),
                static_covariates=series.static_covariates,
                hierarchy=series.hierarchy,
            )
        return series

    def fit(self, series: TimeSeries, future_covariates: Optional[TimeSeries] = None):
        # for VARIMA we need to process target `series` before calling
        # TransferableFutureCovariatesLocalForecastingModel's fit() method
        self._last_values = (
            series.last_values()
        )  # needed for back-transformation when d=1

        series = self._differentiate_series(series)

        super().fit(series, future_covariates)

        return self

    def _fit(
        self, series: TimeSeries, future_covariates: Optional[TimeSeries] = None
    ) -> None:
        super()._fit(series, future_covariates)

        self._assert_multivariate(series)

        # storing to restore the statsmodels model results object
        self.training_historic_future_covariates = future_covariates

        m = staVARMA(
            endog=series.values(copy=False),
            exog=future_covariates.values(copy=False) if future_covariates else None,
            order=(self.p, self.q),
            trend=self.trend,
        )

        self.model = m.fit(disp=0)

    def _predict(
        self,
        n: int,
        series: Optional[TimeSeries] = None,
        historic_future_covariates: Optional[TimeSeries] = None,
        future_covariates: Optional[TimeSeries] = None,
        num_samples: int = 1,
        verbose: bool = False,
    ) -> TimeSeries:

        if num_samples > 1 and self.trend:
            logger.warning(
                "Trends are not well supported yet for getting probabilistic forecasts with ARIMA."
                "If you run into issues, try calling fit() with num_samples=1 or removing the trend from"
                "your model."
            )

        self._last_num_samples = num_samples

        super()._predict(
            n, series, historic_future_covariates, future_covariates, num_samples
        )

        if series is not None:
            self._training_last_values = self._last_values
            # store new _last_values of the new target series
            self._last_values = (
                series.last_values()
            )  # needed for back-transformation when d=1

            series = self._differentiate_series(series)

            # if the series is differentiated, the new len will be = len - 1, we have to adjust the future covariates
            if historic_future_covariates and self.d > 0:
                historic_future_covariates = historic_future_covariates.slice_intersect(
                    series
                )

            # updating statsmodels results object state

            self.model = self.model.apply(
                series.values(copy=False),
                exog=historic_future_covariates.values(copy=False)
                if historic_future_covariates
                else None,
            )

        # forecast before restoring the training state
        if num_samples == 1:
            forecast = self.model.forecast(
                steps=n,
                exog=future_covariates.values(copy=False)
                if future_covariates
                else None,
            )
        else:
            forecast = self.model.simulate(
                nsimulations=n,
                repetitions=num_samples,
                initial_state=self.model.states.predicted[-1, :],
                exog=future_covariates.values(copy=False)
                if future_covariates
                else None,
            )

        forecast = self._invert_transformation(forecast)

        # restoring statsmodels results object state and last values
        if series is not None:
            self.model = self.model.apply(
                self._orig_training_series.values(copy=False),
                exog=self.training_historic_future_covariates.values(copy=False)
                if self.training_historic_future_covariates
                else None,
            )

            self._last_values = self._training_last_values

        return self._build_forecast_series(np.array(forecast))

    def _invert_transformation(self, series_df: pd.DataFrame):
        if self.d == 0:
            return series_df
        if self._last_num_samples > 1:
            series_df = np.tile(
                self._last_values, (self._last_num_samples, 1)
            ).T + series_df.cumsum(axis=0)
        else:
            series_df = self._last_values + series_df.cumsum(axis=0)
        return series_df

    @property
    def min_train_series_length(self) -> int:
        return 30

    def _is_probabilistic(self) -> bool:
        return True

    def _supports_range_index(self) -> bool:
        raise_if(
            self.trend and self.trend != "c",
            "'trend' is not None. Range indexing is not supported in that case.",
            logger,
        )
        return True



from enum import Enum
from typing import Union

import pandas as pd

from darts import TimeSeries
from darts.logging import raise_if_not

# Those freqs can be used to divide Time deltas (the others can't):
DIVISIBLE_FREQS = {"D", "H", "T", "min", "S", "L", "ms", "U", "us", "N"}


class CovariateType(Enum):
    PAST = "past"
    FUTURE_PAST = "future_past"
    HISTORIC_FUTURE = "historic_future"
    FUTURE = "future"
    NONE = None


def _get_matching_index(ts_target: TimeSeries, ts_covariate: TimeSeries, idx: int):
    """
    Given two overlapping series `ts_target` and `ts_covariate` and an index point `idx` of `ts_target`, returns the
    matching index point in `ts_covariate`, based on the ending times of the two series.
    The indices are starting from the end of the series.

    This function is used to jointly slice target and covariate series in datasets. It supports both datetime and
    integer indexed series.

    Note: this function does not check if the matching index value is in `ts_covariate` or not.
    """
    raise_if_not(
        ts_target.freq == ts_covariate.freq,
        "The dataset contains some target/covariates series pair that have incompatible "
        'time axes (not the same "freq") and thus cannot be matched',
    )

    freq = ts_target.freq

    return idx + _index_diff(
        self=ts_target.end_time(), other=ts_covariate.end_time(), freq=freq
    )


def _index_diff(
    self: Union[pd.Timestamp, int], other: Union[pd.Timestamp, int], freq: pd.offsets
):
    """Returns the difference between two indexes `other` and `self` (`other` - `self`) of frequency `freq`."""
    if isinstance(freq, int):
        return int(other - self)

    elif freq.freqstr in DIVISIBLE_FREQS:
        return int((other - self) / freq)

    # /!\ THIS IS TAKING LINEAR TIME IN THE LENGTH OF THE SERIES
    # it won't scale if the end of target and covariates are far apart and the freq is not in DIVISIBLE_FREQS
    # (Not sure there's a way around it for exotic freqs)
    if other >= self:
        return -1 + len(pd.date_range(start=self, end=other, freq=freq))
    else:
        return 1 - len(pd.date_range(start=other, end=self, freq=freq))
